"""Last updated point should not be longer ago than a user-specified number of seconds.
"""

from datetime import timedelta

from timeseer import AnalysisInput, AnalysisResult, EventFrame

_CHECK_NAME = "Data older than threshold"
_EVENT_FRAME_NAME = "Data older than threshold"

META: dict = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
            "data_type": "bool",
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
        }
    ],
    "signature": "univariate",
}


def _is_last_point_old(analysis_input: AnalysisInput) -> bool:
    last_recorded_time = analysis_input.data.index[-1]
    last_evaluation_date = analysis_input.evaluation_time_range.end_date

    threshold = analysis_input.parameters["threshold"]
    cutoff_date = last_evaluation_date - timedelta(seconds=threshold)
    return last_recorded_time <= cutoff_date


def _is_valid_input(analysis_input: AnalysisInput) -> tuple[str, bool]:
    if len(analysis_input.data) == 0:
        return "No data", False
    if "threshold" not in analysis_input.parameters:
        return "No threshold parameter provided", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    message, is_ok = _is_valid_input(analysis_input)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    is_too_old = _is_last_point_old(analysis_input)
    event_frames = []
    if is_too_old is True:
        event_frames.append(
            EventFrame(
                type=_EVENT_FRAME_NAME,
                start_date=analysis_input.data.index[-1].to_pydatetime(),
                end_date=None,
            )
        )
    return AnalysisResult(event_frames=event_frames)



"""Identification of spikes in consecutive values.
"""

from datetime import timedelta

import pandas as pd
import numpy as np

from timeseer import AnalysisInput, AnalysisResult

from timeseer import DataType
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    process_open_intervals,
    handle_open_intervals,
)

META = {
    "checks": [
        {
            "name": "Spikes (upwards)",
            "event_frames": ["Spike outlier (upwards)"],
        },
        {
            "name": "Spikes (downwards)",
            "event_frames": ["Spike outlier (downwards)"],
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _get_intervals(outliers, df, event_type):
    if outliers is None:
        return pd.DataFrame()
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _spread_active_points(active_series):
    ilocs = np.array(
        [active_series.index.get_loc(ind) for ind in active_series[active_series].index]
    )
    neighbor_ilocs = np.concatenate((ilocs - 1, ilocs + 1, ilocs), axis=None)
    result_series = active_series.copy()
    result_series.iloc[neighbor_ilocs] = True
    return result_series


def _is_valid_input(analysis_input) -> tuple[str, bool]:
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No clean data", False
    if not np.issubdtype(_clean_dataframe(analysis_input.data)["value"], float):
        return "Invalid data type", False
    return "OK", True


# pylint: disable=too-many-arguments
def _find_spikes(
    df,
    median_archival_step,
    isolation_threshold,
    asymmetry_threshold,
    window_size,
    strictness,
):
    time_threshold = timedelta(seconds=(median_archival_step * isolation_threshold))
    df["diff_B"] = df["value"].diff().bfill()
    df["diff_F"] = df["value"].diff().shift(-1).ffill()
    df["ts_diff_B"] = df.index.to_series().diff().bfill()
    df["ts_diff_F"] = df.index.to_series().diff().shift(-1).ffill()
    df["diff_abs"] = df["diff_B"].abs()
    df["q25"] = (
        df["diff_abs"].rolling(window_size, min_periods=1, center=True).quantile(0.25)
    )
    df["q75"] = (
        df["diff_abs"].rolling(window_size, min_periods=1, center=True).quantile(0.75)
    )
    df["iqr"] = df["q75"] - df["q25"]
    df["upper bound"] = df["q75"] + strictness * df["iqr"]
    df["peak"] = (
        (df["diff_B"] * df["diff_F"] < 0)
        & (df["diff_B"] > df["upper bound"])
        & (df["diff_F"] < -df["upper bound"])
    )
    df["valley"] = (
        (df["diff_B"] * df["diff_F"] < 0)
        & (df["diff_B"] < -df["upper bound"])
        & (df["diff_F"] > df["upper bound"])
    )
    df["spike"] = (df["peak"].isin([True])) | (df["valley"].isin([True]))
    df["symmetrical"] = (
        (df["spike"].isin([True]))
        & (abs(df["diff_F"] / df["diff_B"]) < (1 + asymmetry_threshold))
        & (abs(df["diff_F"] / df["diff_B"]) > (1 - asymmetry_threshold))
    )
    df["zigzag"] = (
        (df["spike"].isin([True])) & (df["spike"].shift(-1).isin([True]))
    ) | ((df["spike"].isin([True])) & (df["spike"].shift(1).isin([True])))
    df["isolated_spike"] = (
        (df["spike"].isin([True]))
        & (df["ts_diff_B"] > time_threshold)
        & (df["ts_diff_F"] > time_threshold)
    )
    df["up_spikes"] = (
        (df["peak"].isin([True]))
        & (df["symmetrical"].isin([True]))
        & (df["zigzag"].isin([False]))
        & (df["isolated_spike"].isin([False]))
    )
    df["down_spikes"] = (
        (df["valley"].isin([True]))
        & (df["symmetrical"].isin([True]))
        & (df["zigzag"].isin([False]))
        & (df["isolated_spike"].isin([False]))
    )
    return df["up_spikes"], df["down_spikes"]


# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def _run_spikes_check(
    analysis_input: AnalysisInput,
    median_archival_step,
    isolation_threshold=5,
    asymmetry_threshold=0.2,
    window_size=300,
    strictness=3,
):
    df = _clean_dataframe(analysis_input.data)

    active_points_upwards, active_points_downwards = _find_spikes(
        df,
        median_archival_step,
        isolation_threshold,
        asymmetry_threshold,
        window_size,
        strictness,
    )

    active_points_upwards = _spread_active_points(active_points_upwards)
    active_points_downwards = _spread_active_points(active_points_downwards)
    df = _clean_dataframe(analysis_input.data)

    intervals_downwards = _get_intervals(
        active_points_downwards, df, "Spike outlier (downwards)"
    )
    intervals_downwards = handle_open_intervals(df, intervals_downwards)

    intervals_upwards = _get_intervals(
        active_points_upwards, df, "Spike outlier (upwards)"
    )
    intervals_upwards = handle_open_intervals(df, intervals_upwards)

    all_intervals = pd.concat([intervals_downwards, intervals_upwards])

    frames = event_frames_from_dataframe(process_open_intervals(all_intervals))

    last_analyzed_point = df.index[-2]

    return list(frames), last_analyzed_point


def _get_relevant_statistic(analysis_input: AnalysisInput, stat_name: str):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return statistics[0]


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    message, is_ok = _is_valid_input(analysis_input)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    median_archival_step = _get_relevant_statistic(
        analysis_input, "Archival time median"
    )

    if median_archival_step is None:
        return AnalysisResult(condition_message="No median archival step")

    event_frames, last_analyzed_point = _run_spikes_check(
        analysis_input, median_archival_step
    )

    return AnalysisResult(
        event_frames=event_frames,
        last_analyzed_point=last_analyzed_point,
    )



"""Anomaly based on persist.

<p>Anomaly detection based on a significant difference in a value compared to a
previous window.</p>
<p><img src='../static/images/reporting/persist.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where a persist anomaly occurs. Imagine that 100 points are analyzed in a given time-frame
and there are 10 persist anomalies. The score for this check in that case would be
90% = 1 - 10 / 100. Which means that for 90% of all points no persist anomaly occurs.</p>
<div class="ts-check-impact">
<p>
Outliers in series can cloud downstream analytics and put the burden of removing these without
background knowledge on the algorithms instead of those in the know.
Several calculations are sensitive to these type of sudden changes and process decision chains
based on such an outlier can propagate through the system.
Any type of anomaly detection based on normal behavior of the phyiscal sensor can act as a guide for
prioritization of callibration / maintenance.
</p>
</div>
"""

import logging

from datetime import datetime

import pandas as pd
import numpy as np

from adtk.detector import PersistAD
from pandas.api.types import is_string_dtype

from timeseer import (
    AnalysisInput,
    AnalysisResult,
    DataType,
    EventFrame,
    InterpolationType,
)
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    process_open_intervals,
    handle_open_intervals,
)

_CHECK_NAME = "Persist Anomaly"

META: dict = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": ["Persist Anomaly"],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
            "interpolation_type": [None, InterpolationType.LINEAR],
        }
    ],
    "signature": "univariate",
}

logger = logging.getLogger(__name__)


def _get_intervals(active_points, df: pd.DataFrame, event_type: str):
    interval_grp = (active_points != active_points.shift().bfill()).cumsum()
    active_points[active_points.isna()] = 0
    active_points = np.array(active_points, dtype=bool)
    intervals = (
        df.assign(interval_grp=interval_grp)[active_points]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = event_type
    return intervals


def _get_last_analyzed_point(df: pd.DataFrame, window_size: int) -> datetime:
    if len(df) < window_size:
        return df.index[0]
    last_index = -2 * window_size
    return df.index[last_index]


def _clean_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_persist_ad(
    analysis_input: AnalysisInput,
) -> tuple[list[EventFrame], datetime]:
    df = _clean_dataframe(analysis_input.data)
    window_size = 10
    if len(df["value"]) <= window_size:
        return [], df.index[0]

    persist_ad = PersistAD(window=10)
    try:
        active_points = persist_ad.fit_detect(df["value"])
    except (ValueError, RuntimeError) as err:
        logger.error("Error in persist anomaly: %s", err)
        return [], df.index[0]

    intervals = _get_intervals(active_points, df, "Persist Anomaly")
    intervals = handle_open_intervals(df, intervals)

    frames = event_frames_from_dataframe(process_open_intervals(intervals))

    last_analyzed_point = _get_last_analyzed_point(df, window_size)

    return list(frames), last_analyzed_point


def _is_valid_input(analysis_input: AnalysisInput) -> tuple[str, bool]:
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No clean data", False
    if is_string_dtype(analysis_input.data["value"]):
        return "Can not be string", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(analysis_input: AnalysisInput) -> AnalysisResult:
    message, is_ok = _is_valid_input(analysis_input)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]
    if median_archival_step is None or len(median_archival_step) == 0:
        return AnalysisResult(condition_message="No median archival step")

    frames, last_analyzed_point = _run_persist_ad(analysis_input)

    return AnalysisResult(event_frames=frames, last_analyzed_point=last_analyzed_point)



"""Interpolation information loss for a series at different frequencies.

<p>The Interpolation Information Loss is a measure to indicate how much
of the variance is lost when interpolating time series at a particular frequency.
It is well known that just resampling at fixed intervals can cause missing extremes
and hence create data that is no longer representative for the system.</p>"""

from datetime import timedelta

import pandas as pd
import numpy as np

from pandas.api.types import is_string_dtype

import timeseer

from timeseer import DataType
from timeseer.metadata import fields


META = {
    "statistics": [],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


freq_map = pd.DataFrame(
    {
        "keys": ["1s", "5s", "15s", "1m", "5m", "15m", "1h", "12h", "24h"],
        "values": [1, 5, 15, 60, 300, 900, 3600, 3600 * 12, 3600 * 24],
    }
)


def _get_bar_losses(ii_losses):
    values = [(loss["key"], loss["value"]) for loss in ii_losses]
    statistic = timeseer.Statistic("Interpolation information loss", "table", values)
    return statistic


def _get_statistics(ii_losses):
    statistics = [
        timeseer.Statistic(
            "Interpolation information loss (" + loss["key"] + ")",
            "hidden",
            loss["value"],
        )
        for loss in ii_losses
    ]
    return statistics


def _get_relevant_frequencies(median_archival_step):
    return freq_map[freq_map["values"] >= median_archival_step]


def _get_interpolation_style(analysis_input: timeseer.AnalysisInput) -> str:
    interpolation_type = analysis_input.metadata.get_field(fields.InterpolationType)
    if interpolation_type is None:
        return "time"
    if interpolation_type == "STEPPED":
        return "pad"
    return "time"


def _median_diff_times(df):
    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()
    diff_times.dropna()
    diff_times = diff_times[1:].apply(timedelta.total_seconds)
    return np.median(diff_times)


def _is_trivial(df):
    if df["value"].max() == df["value"].min():
        return True
    if all(df["value"].isna()):
        return True
    return False


def _clean_dataframe(df):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _get_iqr(df):
    q1, q9 = np.nanquantile(df, [0.1, 0.9])
    return q9 - q1


def _calculate_distance(raw_df, sampled_df):
    df = pd.DataFrame({"raw": raw_df["value"], "sampled": sampled_df["value"]})
    df = df.interpolate("time")
    point_distance = df["raw"] - df["sampled"]
    return (
        np.nansum(abs(point_distance))
        / point_distance.notnull().sum()
        / _get_iqr(raw_df)
        * 100
    )


def _get_interpolation_information_loss(
    raw_df: pd.DataFrame, median_archival_step, interpolation_style="time"
):
    raw_df = _clean_dataframe(raw_df)
    if _is_trivial(raw_df):
        return None

    frequencies = _get_relevant_frequencies(median_archival_step)
    if frequencies is None:
        return None

    losses = []
    for _, row in frequencies.iterrows():
        sampled_df = (
            raw_df.resample(rule=timedelta(seconds=row["values"]))
            .mean()
            .interpolate(interpolation_style)
        )
        losses.append(
            {"key": row["keys"], "value": _calculate_distance(raw_df, sampled_df)}
        )
    return losses


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    if is_string_dtype(analysis_input.data["value"]):
        return timeseer.AnalysisResult(condition_message="Can not be a string")
    interpolation_style = _get_interpolation_style(analysis_input)
    median_archival_step = _median_diff_times(analysis_input.data)

    ii_losses = _get_interpolation_information_loss(
        analysis_input.data, median_archival_step, interpolation_style
    )
    if ii_losses is None:
        return timeseer.AnalysisResult(
            condition_message="No interpolation information loss"
        )
    statistics = _get_statistics(ii_losses)
    statistics.append(_get_bar_losses(ii_losses))

    return timeseer.AnalysisResult(statistics=statistics)



"""Implementation details for the Timeseer Client.

Only use classes and functions defined in timeseer_client."""

from enum import Enum


class AugmentationStrategy(Enum):
    """AugmentationStrategy dictates what happens to filtered data points while augmenting based on event frames."""

    REMOVE = "remove values"
    HOLD_LAST = "hold last value"
    LINEAR_INTERPOLATION = "linear interpolation"
    KNN_IMPUTATION = "knn imputation"
    MEAN = "mean"


class TimeseerClientException(Exception):
    """Base class for Timeseer client exceptions.

    Use this to catch any exception that originates in the client."""


class AugmentationException(TimeseerClientException):
    """Exception raised when augmentation strategy fails."""


class UnknownAugmentationStrategyException(TimeseerClientException):
    """Raised when the augmentation strategy is not known."""


class MissingModuleException(TimeseerClientException):
    """Raised when a required Python module is not available."""

    def __init__(self, module_name: str):
        TimeseerClientException.__init__(
            self,
            f'missing Python package: "{module_name}"',
        )


class ProcessType(Enum):
    """ProcessType represents the process type of a time series."""

    CONTINUOUS = "CONTINUOUS"
    REGIME = "REGIME"
    BATCH = "BATCH"
    COUNTER = "COUNTER"



""" Check if the unit has a dimension that matches flow"""


from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.analysis.utils.unit_dimension_match import unit_not_matching_dimension


_CHECK_NAME = "flow unit dimension"
_EVENT_FRAME_NAME = "Unit not matching flow"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
            "data_type": [DataType.STRING],
        }
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:

    parameters = {}
    parameters["value"] = "[mass]/[time]"
    analysis_input.parameters = parameters
    mass_results = unit_not_matching_dimension.run(analysis_input)
    parameters["value"] = "[volume]/[time]"
    analysis_input.parameters = parameters
    volume_results = unit_not_matching_dimension.run(analysis_input)
    check_fail = bool(len(mass_results.event_frames) * len(volume_results.event_frames))
    event_frames = []
    if check_fail:
        event_frames.append(
            EventFrame(
                type=_EVENT_FRAME_NAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=None,
            )
        )

    return AnalysisResult(event_frames=event_frames)



"""
SP Anomalies
"""

from datetime import timedelta
from typing import Optional


import pandas as pd
import numpy as np

from pandas.api.types import is_string_dtype

from timeseer import (
    AnalysisInput,
    AnalysisResult,
    BivariateCheckResult,
    DataType,
    EventFrame,
    MultivariateAnalysisInput,
)

from timeseer.analysis.utils import (
    event_frames_from_dataframe,
)


_CHECK_NAME = "SP Anomalies"
_EVENT_FRAME_NAME = "SP Anomaly"
_MIN_SERIES = 2

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": _MIN_SERIES,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64, DataType.CATEGORICAL],
        }
    ],
    "signature": "bivariate",
}


def _sort_positions(
    inputs: list[AnalysisInput],
) -> tuple[dict, dict]:
    df_dict = {}
    names_dict = {}
    for column in inputs:
        position = column.metadata.get_field_by_name("control loop")
        df = column.data
        df = df.rename({"value": position}, axis="columns")
        df_dict[position] = df
        names_dict[position] = column.metadata.series
    return df_dict, names_dict


def _match_ts_merge(df_dict: dict) -> Optional[pd.DataFrame]:
    if "PV" not in df_dict or "SP" not in df_dict:
        return None
    df = pd.concat([df_dict["PV"], df_dict["SP"]], axis=1)
    df["SP"] = df["SP"].interpolate(method="pad").bfill()
    df.dropna(subset=["PV"], inplace=True)
    return df


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _get_sp_events(df: pd.DataFrame):
    change_points = ~(df["SP"].eq(df["SP"].shift().bfill()))
    sp_events = df.iloc[0, :].to_frame().transpose()
    sp_events = pd.concat(
        [sp_events, df[change_points], df.iloc[-1, :].to_frame().transpose()]
    )
    return list(sp_events.index)


def _get_active_points(df: pd.DataFrame):
    sp_events = _get_sp_events(df)
    median_archival_step = _get_mv_median_archival_step(df)
    active_points = pd.Series(index=df.index, dtype=bool)
    for i in range(1, len(sp_events)):
        segment_length = sp_events[i] - sp_events[i - 1]
        if segment_length.total_seconds() >= 100 * median_archival_step:
            active_points.loc[sp_events[i - 1] : sp_events[i]] = False
    return active_points


def _get_intervals(df, anomalies_mse, event_type):
    anomalies = pd.Series(data=anomalies_mse, index=df.index)
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()

    intervals = (
        df.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = event_type
    return intervals


def _filter_invalid_inputs(
    inputs: list[AnalysisInput],
) -> list[AnalysisInput]:
    valid_inputs = []
    for check_input in inputs:
        if is_string_dtype(check_input.data["value"]):
            continue
        valid_inputs.append(check_input)
    return valid_inputs


def _get_mv_median_archival_step(df):
    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()
    diff_times.dropna()
    return pd.Timedelta(np.median(diff_times)).total_seconds()


def _calculate_total_check_active_time(frames, median_archival_step):
    active_time = timedelta(0)
    for _, row in frames.iterrows():
        length = row["end_date"] - row["start_date"]
        if length == timedelta(0):
            length = timedelta(seconds=median_archival_step)
        active_time = active_time + length
    return active_time


def _run_sp_anomalies_detection(
    inputs: list[AnalysisInput],
) -> tuple[list[EventFrame], dict]:
    df_dict, names_dict = _sort_positions(inputs)
    df = _match_ts_merge(df_dict)
    if df is None:
        return [], {}

    df = _clean_dataframe(df)

    active_points = _get_active_points(df)
    intervals = _get_intervals(df, active_points, _EVENT_FRAME_NAME)

    frames = event_frames_from_dataframe(intervals)

    return list(frames), names_dict


def run(  # pylint: disable=missing-function-docstring
    analysis_input: MultivariateAnalysisInput,
) -> AnalysisResult:
    inputs = _filter_invalid_inputs(analysis_input.inputs)
    if len(inputs) < _MIN_SERIES:
        return AnalysisResult()

    event_frames, names_dict = _run_sp_anomalies_detection(inputs)
    results = [
        BivariateCheckResult(
            _CHECK_NAME,
            names_dict["PV"],
            names_dict["SP"],
            event_frames=event_frames,
        )
    ]

    return AnalysisResult(bivariate_check_results=results)



"""There is no change in the data for a period longer than expected based on history.
"""

import re

from timeseer import AnalysisInput, AnalysisResult, EventFrame


_CHECK_NAME = "Series names match convention"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": ["Series names match convention"],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
        }
    ],
    "signature": "univariate",
}


def _convention_fails(name, convention) -> bool:
    patterns = [re.compile(pattern) for pattern in convention]
    if any(re.fullmatch(pattern, name) for pattern in patterns):
        return False
    return True


def _run_series_names_match_convention(
    analysis_input: AnalysisInput,
) -> list[EventFrame]:
    name = analysis_input.metadata.series.name
    convention = analysis_input.parameters["convention"]

    event_frames: list[EventFrame] = []
    if _convention_fails(name, convention):
        event_frames.append(
            EventFrame(
                type=_CHECK_NAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=None,
            )
        )
    return event_frames


def _convention_is_regex_list(convention: list[str]) -> bool:
    try:
        [re.compile(pattern) for pattern in convention]
    except re.error:
        return False
    return True


def _is_valid_input(analysis_input: AnalysisInput) -> tuple[str, bool]:
    if "convention" not in analysis_input.parameters:
        return "No convention parameter provided", False
    if not _convention_is_regex_list(analysis_input.parameters["convention"]):
        return "Convention is not a valid regex list", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(analysis_input: AnalysisInput) -> AnalysisResult:
    message, is_ok = _is_valid_input(analysis_input)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    event_frames = _run_series_names_match_convention(analysis_input)

    return AnalysisResult(event_frames=event_frames)



"""Multivariate correlation structure without outliers.

<p>This depicts the correlation between all tags in the flow over the time frame of the flow.
Outliers have been removed for this correlation calculation.
Outliers include: all zero measurements and NaNs</p>"""

import pandas as pd

from pandas.api.types import is_string_dtype

import timeseer

from timeseer import DataType


_MIN_SERIES = 2

META: dict = {
    "run": "before",
    "statistics": [
        {"name": "Correlation structure"},
    ],
    "conditions": [
        {
            "min_series": _MIN_SERIES,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [
                DataType.FLOAT32,
                DataType.FLOAT64,
                DataType.DICTIONARY,
                DataType.CATEGORICAL,
            ],
        }
    ],
    "signature": "multivariate",
}


def _rename_to_index(data: pd.DataFrame, index: int) -> pd.DataFrame:
    return data.rename({"value": index}, axis=1, copy=False)[index]


def _clean_inputs(inputs):
    concatenated_df = (
        pd.concat(
            [
                _rename_to_index(series.data, index)[
                    ~series.data.index.duplicated(keep="first")
                ]
                for index, series in enumerate(inputs)
            ],
            axis=1,
            sort=False,
        )
        .interpolate("time")
        .dropna()
    )
    concatenated_df = concatenated_df[(concatenated_df != 0).all(1)]
    return concatenated_df


def _run_multivariate_correlation(inputs):
    clean_df = _clean_inputs(inputs)
    if len(clean_df) < 30:
        return None

    return clean_df.corr()


def _filter_invalid_inputs(
    inputs: list[timeseer.AnalysisInput],
) -> list[timeseer.AnalysisInput]:
    valid_inputs = []
    for check_input in inputs:
        if is_string_dtype(check_input.data["value"]):
            continue
        valid_inputs.append(check_input)
    return valid_inputs


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.MultivariateAnalysisInput,
) -> timeseer.AnalysisResult:
    inputs = _filter_invalid_inputs(analysis_input.inputs)
    if len(inputs) < _MIN_SERIES:
        return timeseer.AnalysisResult()

    correlation = _run_multivariate_correlation(inputs)
    if correlation is None:
        return timeseer.AnalysisResult()

    correlation_dict = correlation.to_dict("list")
    result = {
        "series": [inputs[index].metadata.series for index in correlation_dict],
        "columns": list(correlation_dict.values()),
    }
    correlation_statistic = timeseer.Statistic(
        META["statistics"][0]["name"], "correlation", result
    )
    return timeseer.AnalysisResult(statistics=[correlation_statistic])



"""Utility functions for unit conversions."""

import re
from pathlib import Path
import numpy as np
from pint import UnitRegistry, UndefinedUnitError, DefinitionSyntaxError


def get_unit_registry() -> UnitRegistry:
    """Return a pint UnitRegistry configured for Timeseer."""
    return UnitRegistry(str(Path(__file__).parent.resolve()) + "/default_replace.txt")


def clean_unit_string(unit: str) -> str:
    """Clean up the unit value for parsing in pint."""
    return re.sub(r"\s+", "_", unit.strip())


def get_dimension(unit):
    """Get dimensionality using pint library"""
    ureg = get_unit_registry()
    try:
        return ureg.get_dimensionality(clean_unit_string(unit))
    except (
        UndefinedUnitError,
        AttributeError,
        TypeError,
        ValueError,
        KeyError,
        DefinitionSyntaxError,
    ):
        return "Unknown"


def get_symbol(unit):
    """Get symbol using pint library"""
    ureg = get_unit_registry()
    try:
        return ureg.get_symbol(clean_unit_string(unit))
    except (
        UndefinedUnitError,
        AttributeError,
        TypeError,
        ValueError,
        KeyError,
        DefinitionSyntaxError,
    ):
        return "Unknown"


def get_limited_units():
    """List of units and dimensions where sensible limits are known"""
    ureg = get_unit_registry()
    units = [
        "hertz",
    ]
    dimensions = [
        # "meter",
        "kelvin",
        "second",
        # "ampere",
        "candela",
        "gram",
        "mole",
        "unit",
        "pH",
        "m2",
        "liter",
        # "kph",
        # "galileo",
        "newton",
        "joule",
        "watt",
        "water",
        # "pascal",
        "foot_pound",
        "poise",
        "stokes",
        "rhe",
        "particle",
        "molar",
        "katal",
        "clausius",
        "entropy_unit",
        # "curie",
        "langley",
        "nit",
        "lumen",
        "lux",
        "a_u_intensity",
        # "volt",
        "ohm",
        "siemens",
        "henry",
        "weber",
        "tesla",
        # "bohr_magneton",
    ]
    units_zero_min = [ureg.get_symbol(unit) for unit in units]
    dimensions_zero_min = [
        ureg.get_dimensionality(dimension) for dimension in dimensions
    ]
    return units_zero_min, dimensions_zero_min


def get_unit_range(unit):
    """Returns min and max value for general case of units with known limits
    and considering the special cases."""
    min_value = 0
    max_value = np.inf
    if get_dimension(unit) == get_dimension("pH"):
        max_value = 14
    if get_symbol(unit) == get_symbol("C"):
        min_value = -273.15
    if get_symbol(unit) == get_symbol("F"):
        min_value = -459.67
    return min_value, max_value



""" Calculate the distributed quantile sketch of the time series. """

import jsonpickle

from ddsketch.ddsketch import DDSketch
from pandas.api.types import is_string_dtype

import timeseer

from timeseer import DataType


META: dict = {
    "statistics": [
        {"name": "Value Sketch"},
    ],
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64, None],
        }
    ],
    "signature": "univariate",
}


def _clean_dataframe(df):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_get_sketch(analysis_input, previous_sketch):
    df = _clean_dataframe(analysis_input.data)
    if len(df) == 0:
        return previous_sketch

    if previous_sketch is None:
        previous_sketch = DDSketch(0.001)

    for v in df["value"]:
        previous_sketch.add(v)

    return previous_sketch


def _is_valid_input(analysis_input: timeseer.AnalysisInput) -> tuple[str, bool]:
    if is_string_dtype(analysis_input.data["value"]):
        return "Can not be a string", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    message, is_ok = _is_valid_input(analysis_input)
    if not is_ok:
        return timeseer.AnalysisResult(condition_message=message)

    previous_sketch = None
    relevant_statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == META["statistics"][0]["name"]
    ]
    if relevant_statistics is not None and len(relevant_statistics) != 0:
        previous_sketch = jsonpickle.decode(relevant_statistics[0])

    sketch = _run_get_sketch(analysis_input, previous_sketch)
    if sketch is None:
        return timeseer.AnalysisResult(condition_message="No sketch")

    sketch_statistic = timeseer.Statistic(
        META["statistics"][0]["name"], "sketch", jsonpickle.encode(sketch)
    )
    return timeseer.AnalysisResult(statistics=[sketch_statistic])



"""Identification of sensor drift between 2 series by means of cusum.

<p>This check identifies periods where for redundant sensors values of one of the
sensors starts to drift away from the other one.</p>
<p><img src='../static/images/reporting/sensor_drift.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where sensor drift is identified. Imagine that 100 points are analyzed in a given time-frame
and that drift is detected for 10 (consecutive) points. The score for this check in that case would be
90% = 1 - 10 / 100. Which means that for 90% of all points no drift occurs.</p>
<div class="ts-check-impact">
<p>Changes in a physical relation between a set of series could indicate process or instrumentation issues.</p>

</div>
"""

import pandas as pd
import numpy as np

from timeseer import (
    AnalysisInput,
    AnalysisResult,
    BivariateCheckResult,
    DataType,
    MultivariateAnalysisInput,
)
from timeseer.analysis.utils import event_frames_from_dataframe


_CHECK_NAME = "Redundant sensor drift"
_EVENT_FRAME_NAME = "Redundant sensor drift"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": 2,
            "max_series": 2,
        },
        {
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        },
    ],
    "signature": "bivariate",
}


def cusum(
    drift_series, threshold, anomaly_threshold
):  # pylint: disable=missing-function-docstring
    series = np.array(drift_series["drifts"], dtype=float)
    series = series / np.nanstd(series)

    positive_cusum = (series * 0).copy()
    negative_cusum = positive_cusum.copy()
    for i in np.arange(len(series)):
        if i == 0:
            positive_cusum[i] = series[i]
            negative_cusum[i] = series[i]
        else:
            positive_cusum[i] = np.max(
                [0, series[i] - threshold + positive_cusum[i - 1]]
            )
            negative_cusum[i] = np.max(
                [0, -threshold - series[i] + negative_cusum[i - 1]]
            )
    anomalies = (positive_cusum > anomaly_threshold) | (
        negative_cusum > anomaly_threshold
    )
    return anomalies


def _get_intervals(anomalies, df):
    anomalies = pd.Series(data=anomalies, index=df.index)
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()

    intervals = (
        df.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = _EVENT_FRAME_NAME
    return intervals


def _get_result_per_serie(serie, df) -> BivariateCheckResult:
    anomalies = cusum(serie, 0.5, 3)
    intervals = _get_intervals(anomalies, df)

    frames = event_frames_from_dataframe(intervals)

    return BivariateCheckResult(
        _CHECK_NAME,
        serie["series_x"],
        serie["series_y"],
        event_frames=list(frames),
    )


def _get_individual_difference_series(diff_matrix, series):
    results = []
    for i in range(len(series) - 1):
        for j in range(i + 1, len(series)):
            results.append(
                {
                    "series_x": series[i],
                    "series_y": series[j],
                    "drifts": [difference[i, j] for difference in diff_matrix],
                }
            )
    return results


def _make_diff_matrix(df):
    diff_matrix = np.zeros((df.shape[0], df.shape[1], df.shape[1]))
    for i in range(df.shape[1] - 1):
        for j in range(i + 1, df.shape[1]):
            diff_matrix[:, i, j] = df.iloc[:, i] - df.iloc[:, j]
    return diff_matrix


def _clean_input(inputs, selected_index):
    index = inputs[selected_index].data.index
    return (
        pd.concat(
            [
                series.data["value"][
                    ~series.data["value"].index.duplicated(keep="first")
                ]
                for series in inputs
            ],
            axis=1,
            sort=False,
        )
        .reindex(index)
        .interpolate("time")
        .dropna()
        .sort_index()
    )


def _get_median_archival_step(df):
    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()
    diff_times.dropna()
    return pd.Timedelta(np.median(diff_times)).total_seconds()


def _run_detect_redundant_sensor_cusum(
    inputs: list[AnalysisInput],
) -> list[BivariateCheckResult]:
    median_archival_step1 = _get_median_archival_step(inputs[0].data)
    median_archival_step2 = _get_median_archival_step(inputs[1].data)
    selected_index = 1
    if median_archival_step1 > median_archival_step2:
        selected_index = 0
    concatenated_df = _clean_input(inputs, selected_index)

    diff_matrix = _make_diff_matrix(concatenated_df)

    series = [series.metadata.series for series in inputs]
    individual_difference_series = _get_individual_difference_series(
        diff_matrix, series
    )

    return [
        _get_result_per_serie(serie, concatenated_df)
        for serie in individual_difference_series
    ]


def run(
    analysis_input: MultivariateAnalysisInput,
):  # pylint: disable=missing-function-docstring
    inputs = analysis_input.inputs

    results = _run_detect_redundant_sensor_cusum(inputs)

    return AnalysisResult(bivariate_check_results=results)



"""For certain units (e.g. pH) there is a limited range values can take.

<p>This check validates whether for a specific set of units, the historical values are
within the expected deterministic range.</p>
<p class="scoring-explanation">The score for this check is based on the amount
of time values are outside of the unit range.</p>
<div class="ts-check-impact">
<p>Mismatches between unit range and measured values are data quality issues.
</p>
</div>
"""

from enum import Enum, auto

import pandas as pd

from pandas.api.types import is_string_dtype

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.analysis.utils import (
    get_dimension,
    get_symbol,
    get_limited_units,
    get_unit_range,
    event_frames_from_dataframe,
    process_open_intervals,
    handle_open_intervals,
)
from timeseer.metadata import fields

_CHECK_NAME = "Values outside unit range"
_EVENT_FRAME_NAME = "Values outside unit range"


META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


class LimitType(Enum):
    """Choose whether to use physical or functional limits."""

    PHYSICAL = auto()
    FUNCTIONAL = auto()


def _get_intervals(outliers, df, event_type):
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _get_active_points(df: pd.DataFrame, min_value: float, max_value: float):
    return (df["value"] < min_value) | (df["value"] > max_value)


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_values_outside_unit_range_check(
    unit, analysis_input: AnalysisInput
) -> list[EventFrame]:
    df = analysis_input.data
    df = _clean_dataframe(df)

    min_value, max_value = get_unit_range(unit)

    active_points = _get_active_points(df, min_value, max_value)
    intervals = _get_intervals(active_points, df, _EVENT_FRAME_NAME)
    intervals = handle_open_intervals(df, intervals)

    frames = event_frames_from_dataframe(process_open_intervals(intervals))

    return list(frames)


def _is_valid_input(analysis_input: AnalysisInput, unit: str) -> tuple[str, bool]:
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No data", False
    if is_string_dtype(analysis_input.data["value"]):
        return "Can not be a string", False
    if unit is None:
        return "No unit", False
    units_zero_min, dimensions_zero_min = get_limited_units()
    if (get_dimension(unit) not in dimensions_zero_min) & (
        get_symbol(unit) not in units_zero_min
    ):
        return "Unit range is unknown", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    unit = analysis_input.metadata.get_field(fields.Unit)

    message, is_ok = _is_valid_input(analysis_input, unit)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    frames = _run_values_outside_unit_range_check(unit, analysis_input)
    return AnalysisResult(
        event_frames=frames, last_analyzed_point=analysis_input.data.index[-1]
    )



"""Time series data should have strict increasing timestamps.

<p>Out of order time stamps can occur because of connection issues, causing the system
to rely on buffered data. Another common cause for out of order samples isthe daylight saving time
not handled correctly in the source.</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>
Misaligned ordering of timestamps can hamper (automated) downstream analytics.
</p>
</div>
"""

from datetime import timedelta

from timeseer import AnalysisInput, AnalysisResult, EventFrame

_CHECK_NAME = "No out-of-order samples"
_EVENT_FRAME = "No out-of-order samples"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME],
            "data_type": "bool",
        }
    ],
    "conditions": [{"min_series": 1, "min_data_points": 2}],
    "signature": "univariate",
}


def _any_out_of_order_samples(df):
    return any(df.index.to_series().diff() < timedelta(0))


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    out_of_order = _any_out_of_order_samples(analysis_input.data)
    event_frames = []
    if out_of_order is True:
        event_frames.append(
            EventFrame(
                type=_EVENT_FRAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=None,
            )
        )
    return AnalysisResult(event_frames=event_frames)



"""Timeseer Client allows querying of data and metadata."""

import json
import time

try:
    import yaml

    HAS_YAML = True
except ImportError:
    HAS_YAML = False


from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Tuple

from kukur import Metadata, SeriesSelector
from kukur.client import Client as KukurClient
import kukur.config

import pyarrow as pa
import pyarrow.flight as fl

from timeseer_client.internal import MissingModuleException


class Client(KukurClient):
    """Client connects to Timeseer using Arrow Flight."""

    def upload_data(
        self,
        metadata_or_data: Union[Metadata, List[Tuple[Metadata, pa.Table]]],
        table: Optional[pa.Table] = None,
        *,
        analyze=True,
        block=True,
    ):
        """Upload time series data to Timeseer.

        This requires a configured 'flight-upload' source in Timeseer.

        There are two ways to call this method.

        One requires two arguments:
            metadata: any known metadata about the time series. This will be merged with the metadata
                already known by Timeseer depending on the source configuration. The source of the series should match
                the source name of a 'flight-upload' source.
            table: a pyarrow.Table of two columns.
                The first column with name 'ts' contains Arrow timestamps.
                The second column with name 'value' contains the values as a number or string.

        The second accepts a list of tuples of the same arguments. This allows uploading multiple time series at the
        same time.

        When `analyze` is `True`, start a flow evaluation.
        When `block` is `True`, block execution until the flow evaluation is done.
        """
        if table is not None:
            assert isinstance(metadata_or_data, Metadata)
            self._upload_data_single(metadata_or_data, table, analyze, block)
        else:
            assert not isinstance(metadata_or_data, Metadata)
            self._upload_data_multiple(metadata_or_data, analyze, block)

    def _upload_data_single(
        self, metadata: Metadata, table: pa.Table, analyze: bool, block: bool
    ):
        self._upload_data_multiple([(metadata, table)], analyze, block)

    def _upload_data_multiple(
        self, many_series: List[Tuple[Metadata, pa.Table]], analyze: bool, block: bool
    ):
        client = self._get_client()
        selectors = []
        for metadata, table in many_series:
            metadata_json = metadata.to_data()
            selector = SeriesSelector.from_tags(
                metadata.series.source, metadata.series.tags, metadata.series.field
            )
            selectors.append(selector)
            descriptor = fl.FlightDescriptor.for_command(
                json.dumps(
                    dict(
                        metadata=metadata_json,
                    )
                )
            )
            writer, reader = client.do_put(descriptor, table.schema)
            writer.write_table(table)
            writer.done_writing()
            buf: pa.Buffer = reader.read()
            response: Dict = json.loads(buf.to_pybytes())
            writer.close()

        if "flowName" in response:
            if analyze:
                _analyze_flow(
                    client, response["flowName"], limitations=selectors, block=block
                )
        elif block and "flow_evaluation_group_id" in response:
            _wait_for_flow_evaluation(client, response)

    def evaluate_flow(self, flow_name: str, *, block=True):
        """Evaluate a flow.

        Args:
            flow_name: the name of the flow to evaluate
            block: block until the evaluation completes (keyword-only, default True)
        """
        client = self._get_client()
        _analyze_flow(client, flow_name, block=block)

    def duplicate_flow(  # pylint:disable=too-many-arguments
        self,
        existing_flow_name: str,
        new_flow_name: str,
        *,
        series_set_names: List[str] = None,
        start_date: datetime = None,
        end_date: datetime = None,
        data_set_name: str = None,
    ):
        """Duplicate an existing flow.

        Args:
            existing_flow_name: the name of the flow to duplicate.
            new_flow_name: the name of the duplicated flow.
            series_set_name: the name of an existing series set to be used in the flow. (Optional).
            start_date: the start date of the flow (Optional).
            end_date: the end date of the flow (Optional).
            data_set_name: the name of the data set to use for the flow (Optional).
        """

        body: Dict = dict(
            existingFlowName=existing_flow_name,
            newFlowName=new_flow_name,
            seriesSetNames=series_set_names,
            dataSetName=data_set_name,
        )
        if start_date is not None:
            body["startDate"] = start_date.isoformat()
        if end_date is not None:
            body["endDate"] = end_date.isoformat()
        self._get_client().do_action(("duplicate_flow", json.dumps(body).encode()))

    def get_event_frames(
        self,
        selector: SeriesSelector,
        start_date: datetime = None,
        end_date: datetime = None,
        frame_type: Union[str, List[str]] = None,
    ) -> pa.Table:
        """Get all event frames matching the given criteria.

        Args:
            selector: the time series source, exposed flow or time series to which the event frames are linked.
            start_date: the start date of the range to find overlapping event frames in. Defaults to one year ago.
            end_date: the end date of the range to find overlapping event frames in. Defaults to now.
            frame_type: the type or types of event frames to search for. Finds all types when empty.

        Returns::
            A pyarrow Table with 6 columns.
            The first column ('start_date') contains the start date.
            The second column ('end_date') contains the end date.
            The third column ('type') contains the type of the returned event frame as a string.
            The fourth column ('explanation') can contain the explanation for an event frame as a string.
            The fifth column ('status') can contain the status of an event frame as a string.
            Columns 6 contains possible multiple references for the event frame.
        """
        if start_date is None or end_date is None:
            now = datetime.utcnow().replace(tzinfo=timezone(timedelta(0)))
            if start_date is None:
                start_date = now.replace(year=now.year - 1)
            if end_date is None:
                end_date = now

        query: Dict[str, Any] = {
            "query": "get_event_frames",
            "selector": selector.to_data(),
            "start_date": start_date.isoformat(),
            "end_date": end_date.isoformat(),
        }

        if frame_type is not None:
            query["type"] = frame_type

        ticket = fl.Ticket(json.dumps(query))
        return self._get_client().do_get(ticket).read_all()

    def get_data_quality_score_data_sources(
        self, source_names: List[str]
    ) -> Dict[str, int]:
        """Get the data quality score of a data source or exposed flow.

        Args:
            source_names: A list of time series sources or Flight Expose blocks

        Returns::
            A data quality score of every given source as a percentage.
        """
        body = {
            "source_names": source_names,
        }
        results = list(
            self._get_client().do_action(
                ("get_data_quality_score_data_sources", json.dumps(body).encode())
            )
        )
        return json.loads(results[0].body.to_pybytes())

    def get_kpi_scores(
        self,
        source_name: str,
    ) -> Dict[str, Dict[str, int]]:
        """Get the kpi scores of a data source or exposed flow.

        Args:
            source_name: The time series source or the name of the Flight Expose block.

        Returns::
            The score per KPI as a percentage, keyed by the source_name provided.
        """
        body = {
            "source_name": source_name,
        }

        results = list(
            self._get_client().do_action(("get_kpi_scores", json.dumps(body).encode()))
        )
        return json.loads(results[0].body.to_pybytes())

    def get_series_results(
        self, selector: SeriesSelector
    ) -> List[Dict[str, Union[str, float]]]:
        """Return the univariate check results for the given series.

        Args:
            selector: The time series to return results for.

        Returns::
            A list of check results, where each check result is a dictionary with 'name' and 'result'.
            The result is returned as a floating point number between 0.0 (bad) and 1.0 (good)."""
        body = {
            "selector": selector.to_data(),
        }
        results = list(
            self._get_client().do_action(
                ("get_series_results", json.dumps(body).encode())
            )
        )
        return json.loads(results[0].body.to_pybytes())

    def get_series_statistics(self, selector: SeriesSelector) -> List[Dict[str, Any]]:
        """Return the univariate statistics for the given series.

        Args:
            selector: The time series to return statistics for.

        Returns::
            A list of statistics. Each statistic is a dict with three fields:
              - name (str)
              - dataType (str) - one of 'float', 'pct', 'datetime', 'hidden', 'table'
              - result (exact type depending on the dataType)"""
        body = {"selector": selector.to_data()}
        results = list(
            self._get_client().do_action(
                ("get_series_statistics", json.dumps(body).encode())
            )
        )
        return json.loads(results[0].body.to_pybytes())

    def remove_data(self, selector: SeriesSelector):
        """Removes the series indicated by the SeriesSelector and related files.
           The source in the selector should match the source name of a
           'flight-upload' source.

        Args:
            selector: The time series to be removed. If selector contains no series
            name then all the series on the given source are removed.
        """
        body = {"selector": selector.to_data()}

        self._get_client().do_action(("remove_series", json.dumps(body).encode()))

    def list_flows(self) -> List[str]:
        """Return a list containing all the flow names."""
        results = list(
            self._get_client().do_action(("get_flows", json.dumps({}).encode()))
        )
        return list(json.loads(results[0].body.to_pybytes()))

    def list_exposed_blocks(self, flow_name: str) -> Dict[str, List[str]]:
        """Return a dictionary containing all the exposed blocks names per series set.

        Args:
            flow_name: The name of the flow.

        Returns::
            The names of the exposed blocks in a flow per series set.
        """

        body = {"flowName": flow_name}

        results = list(
            self._get_client().do_action(
                ("get_exposed_blocks", json.dumps(body).encode())
            )
        )
        return json.loads(results[0].body.to_pybytes())

    def list_data_services(self) -> List[str]:
        """Return a list containing all the data service names."""
        results = list(
            self._get_client().do_action(
                ("list_data_services", json.dumps({}).encode())
            )
        )
        return list(json.loads(results[0].body.to_pybytes()))

    def get_data_service_kpi_scores(
        self, data_service_name: str, view: str
    ) -> Dict[str, int]:
        """Get the kpi scores of a data service.

        Args:
            data_service_name: The name of the Data Service.
            view: The series set name.

        Returns::
            The score per KPI as a percentage, keyed by the series set name included in the data service.
        """
        body = {
            "data_service_name": data_service_name,
            "view": view,
        }

        results = list(
            self._get_client().do_action(
                ("get_data_service_kpi_scores", json.dumps(body).encode())
            )
        )
        return json.loads(results[0].body.to_pybytes())

    def get_data_service_event_frames(  # pylint:disable=too-many-arguments
        self,
        data_service_name: str,
        view: str,
        selector: SeriesSelector = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        frame_type: Union[str, List[str]] = None,
    ) -> pa.Table:
        """Get all event frames matching the given criteria.

        Args:
            data_service_name: The Data Service name to which the event frames are linked.
            view: The series set name to which the event frames are linked
            selector: the time series to which the event frames are linked.
            start_date: the start date of the range to find overlapping event frames in.
                Defaults to start date of the data service.
            end_date: the end date of the range to find overlapping event frames in.
                Defaults to end date of the data service.
            frame_type: the type or types of event frames to search for. Finds all types when empty.

        Returns::
            A pyarrow Table with 6 columns.
            The first column ('start_date') contains the start date.
            The second column ('end_date') contains the end date.
            The third column ('type') contains the type of the returned event frame as a string.
            The fourth column ('explanation') can contain the explanation for an event frame as a string.
            The fifth column ('status') can contain the status of an event frame as a string.
            Columns 6 contains possible multiple references for the event frame.
            The seventh column contains the uuid of the event frame.
        """

        query: Dict[str, Any] = {
            "query": "get_data_service_event_frames",
            "data_service_name": data_service_name,
            "view": view,
            "selector": selector.to_data() if selector is not None else None,
            "start_date": start_date.isoformat() if start_date is not None else None,
            "end_date": end_date.isoformat() if end_date is not None else None,
        }

        if frame_type is not None:
            query["type"] = frame_type

        ticket = fl.Ticket(json.dumps(query))
        return self._get_client().do_get(ticket).read_all()

    def create_resources(
        self,
        resources: List[Dict] = None,
        *,
        resource: Dict = None,
        path: Optional[Union[Path, str]] = None,
    ):
        """Create resources by supplying one resource, multiple resources or a filename.

        Args:
            resources: a list containing the resource definitions.
            resource: One resource definition.
            path: A path to the file or filename of a resource definition.
                This can be in yaml, toml or json format.
        """
        all_resources = []
        if resources is not None:
            all_resources.extend(resources)
        if resource is not None:
            all_resources.append(resource)
        if path is not None:
            if isinstance(path, str):
                path = Path(path)
            if path.suffix in [".yml", ".yaml"]:
                if not HAS_YAML:
                    raise MissingModuleException("PyYAML")
                all_resources.extend(_read_yaml(path))
            elif path.suffix == ".toml":
                all_resources.extend(_read_toml(path))
            elif path.suffix == ".json":
                all_resources.extend(_read_json(path))
        body = dict(
            resources=all_resources,
        )

        self._get_client().do_action(("create_resources", json.dumps(body).encode()))

    def remove_resources(
        self,
        resources: Dict[str, Union[str, List[str]]],
    ):
        """Remove resources by type and name.

        Args:
            resources: A dictionary of the type and the names of the resource to remove.
        """

        body = dict(
            resources=resources,
        )

        self._get_client().do_action(("remove_resources", json.dumps(body).encode()))


def _analyze_flow(
    client: fl.FlightClient,
    flow_name: str,
    *,
    limitations: Optional[List[SeriesSelector]] = None,
    block=True,
):
    flow: Dict = dict(flowName=flow_name)
    if limitations is not None:
        flow["limitations"] = [selector.to_data() for selector in limitations]
    results = list(client.do_action(("evaluate_flow", json.dumps(flow).encode())))
    response = json.loads(results[0].body.to_pybytes())

    if block:
        _wait_for_flow_evaluation(client, response)


def _wait_for_flow_evaluation(client: fl.FlightClient, response: Dict):
    while True:
        results = list(
            client.do_action(
                ("get_flow_evaluation_state", json.dumps(response).encode())
            )
        )
        state = json.loads(results[0].body.to_pybytes())
        if (
            state["completed"] == state["total"]
            and state["blockCompleted"] == state["blockTotal"]
        ):
            break
        time.sleep(1)


def _read_yaml(path: Path) -> List[Dict]:
    with path.open() as file:
        return yaml.safe_load(file)


def _read_toml(path) -> List[Dict]:
    config = kukur.config.from_toml(path)
    return config["resource"]


def _read_json(path: Path) -> List[Dict]:
    with path.open() as file:
        return json.load(file)



"""Lower limit set by the metadata should not be subceeded, taking into account the uncertainty
as defined by the accuracy in the metadata.

<p>Every measurement has an inherent uncertainty given by precision of the measuring
instrument as well as potential compression settings in the historian. So every value
should be interpreted with uncertainty bounds. This check identifies whether the lower limit,
if defined, is crossed taking these uncertainty bounds into account.</p>
<p><img src='../static/images/reporting/limits_accuracy.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where the corresponding value is lower than the given limit, taking into account the accuracy.
Imagine that 100 points are analyzed in a given time-frame
and there are 10 points whose value + accuracy is lower than the given limit.
The score for this check in that case would be
90% = 1 - 10 / 100. Which means that 90% of all points lie above the limit even taken accuracy
into account.</p>
<div class="ts-check-impact">
<p>
When the sensor spec limits are subceeded this is an indication of sensor failure. This might mean the
sensor needs to be recalibrated.
</p>
</div>
"""

import pandas as pd

from pandas.api.types import is_string_dtype

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame, Metadata
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    process_open_intervals,
    handle_open_intervals,
)
from timeseer.metadata import fields

_CHECK_NAME = "Out-of-bounds (lower, physical, accuracy)"
_EVENT_FRAME_NAME = "Out of bounds (lower, physical, accuracy)"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _get_intervals(outliers, df, event_type):
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _get_active_points(df: pd.DataFrame, metadata: Metadata):
    limit_low = metadata.get_field(fields.LimitLowPhysical)
    accuracy = metadata.get_field(fields.Accuracy)
    assert limit_low is not None
    assert accuracy is not None
    return df["value"] < limit_low + accuracy


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_limit_accuracy_check(metadata: Metadata, df: pd.DataFrame) -> list[EventFrame]:
    df = _clean_dataframe(df)

    active_points = _get_active_points(df, metadata)
    intervals = _get_intervals(active_points, df, _EVENT_FRAME_NAME)
    intervals = handle_open_intervals(df, intervals)
    intervals = process_open_intervals(intervals)

    frames = event_frames_from_dataframe(intervals)

    return list(frames)


def _is_input_valid(
    analysis_input: AnalysisInput, median_archival_step: list[float]
) -> tuple[str, bool]:
    if analysis_input.metadata.get_field(fields.Accuracy) is None:
        return "No accuracy", False
    if analysis_input.metadata.get_field(fields.LimitLowPhysical) is None:
        return "No physical lower limit", False
    if median_archival_step is None or len(median_archival_step) == 0:
        return "No median archival step", False
    if is_string_dtype(analysis_input.data["value"]):
        return "Can not be a string", False
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No clean data", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(analysis_input: AnalysisInput) -> AnalysisResult:
    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]
    message, is_ok = _is_input_valid(analysis_input, median_archival_step)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    frames = _run_limit_accuracy_check(analysis_input.metadata, analysis_input.data)
    return AnalysisResult(
        event_frames=frames, last_analyzed_point=analysis_input.data.index[-1]
    )



"""Identification of correlation outliers between two series.

<p>This check identifies periods of behavior in which the correlations between any 2 series
is significantly different from the baseline.</p>
<p><img src='../static/images/reporting/bivariate_correlation_outlier.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the sum of time of all
the event-frames containing correlation outliers. E.g. assume a total period being analyzed of 1 year and
2 event-frames, of 1 month and 2 months respectively.
The score of this check will then be 75% = 1 - 3 / 12.
Which means that in 75% of time no correlation outliers occur.</p>
<div class="ts-check-impact">
<p>A change in the relation between multiple series might indicate potential process upsets,
as well as instrumentation issues.</p>
</div>
"""

from datetime import timedelta

import pandas as pd
import numpy as np

from pandas.api.types import is_string_dtype

from timeseer import (
    AnalysisInput,
    AnalysisResult,
    DataType,
    EventFrame,
    MultivariateAnalysisInput,
)
from timeseer.analysis.utils import event_frames_from_dataframe


_CHECK_NAME = "Correlation outlier"
_EVENT_FRAME_NAME = "Correlation outlier"
_MIN_SERIES = 2

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": _MIN_SERIES,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [
                DataType.FLOAT32,
                DataType.FLOAT64,
                DataType.DICTIONARY,
                DataType.CATEGORICAL,
            ],
        }
    ],
    "signature": "multivariate",
}


def _is_anomaly(lower, upper, arr):
    a_lower = np.any(arr < lower)
    a_upper = np.any(arr > upper)
    return a_lower or a_upper


def _get_active_days(correlations):
    q25, q75 = np.nanquantile(np.array(correlations, dtype=float), [0.25, 0.75], axis=0)
    iqr = q75 - q25
    upper = q75 + 1.5 * iqr
    lower = q25 - 1.5 * iqr
    return [_is_anomaly(lower, upper, x) for x in np.array(correlations, dtype=float)]


def _get_intervals(active_days, days):
    anomalies = pd.Series(data=active_days, index=days.index)
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()

    intervals = (
        days.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["end_date"] = intervals["end_date"] + timedelta(
        hours=23, minutes=59, seconds=59
    )
    intervals["type"] = _EVENT_FRAME_NAME
    return intervals


def _clean_input(inputs):
    concatenated_df = (
        pd.concat(
            [
                series.data[~series.data.index.duplicated(keep="first")]["value"]
                for series in inputs
            ],
            axis=1,
            sort=False,
        )
        .interpolate("time")
        .dropna()
        .sort_index()
    )
    return concatenated_df[(concatenated_df != 0).all(1)]


def _run_correlation_outlier_detection(inputs: list[AnalysisInput]) -> list[EventFrame]:
    concatenated_df = _clean_input(inputs)
    if len(concatenated_df) < 30:
        return []
    daily_correlations = np.array(
        [x[1].corr() for x in concatenated_df.resample("D")], dtype=object
    )

    active_days = _get_active_days(daily_correlations)
    days = concatenated_df.resample("D").mean()
    intervals = _get_intervals(active_days, days)
    frames = event_frames_from_dataframe(intervals)

    return list(frames)


def _filter_invalid_inputs(
    inputs: list[AnalysisInput],
) -> list[AnalysisInput]:
    valid_inputs = []
    for check_input in inputs:
        if is_string_dtype(check_input.data["value"]):
            continue
        valid_inputs.append(check_input)
    return valid_inputs


def run(
    analysis_input: MultivariateAnalysisInput,
):  # pylint: disable=missing-function-docstring
    inputs = _filter_invalid_inputs(analysis_input.inputs)
    if len(inputs) < _MIN_SERIES:
        return AnalysisResult()

    event_frames = _run_correlation_outlier_detection(inputs)

    return AnalysisResult(event_frames=event_frames)



""" Check if the unit has a dimension that matches pressure"""


from timeseer import AnalysisInput, AnalysisResult, DataType
from timeseer.analysis.utils.unit_dimension_match import unit_not_matching_dimension


_CHECK_NAME = "pressure unit dimension"
_EVENT_FRAME_NAME = "Unit not matching pressure"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
            "data_type": [DataType.STRING],
        }
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:

    parameters = {}
    parameters["value"] = "[pressure]"
    analysis_input.parameters = parameters
    return unit_not_matching_dimension.run(analysis_input)



import sqlite3

from yoyo import step
from timeseer import ResourceOrigin

from timeseer.kpi_sets import DefaultKPISet

__depends__ = {"007__change_to_kpi_set"}


def rename_default_kpi_set(db: sqlite3.Connection):
    cursor = db.cursor()
    cursor.execute(
        """
            update KPISets set name = ?, origin = ?
            where name = 'default' and origin = 'default'
        """,
        [DefaultKPISet.DEFAULT.value, ResourceOrigin.DEFAULT.value],
    )


steps = [
    step(rename_default_kpi_set),
]



"""Identification of correlation outliers between each combination of two series.

<p>This check identifies periods of behavior in which the correlation between two series
is significantly different from the baseline.</p>
<p><img src='../static/images/reporting/bivariate_correlation_outlier.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the sum of time of all
the event-frames containing bivariate correlation outliers. E.g. assume a total period being analyzed of 1 year and
2 event-frames, of 1 month and 2 months respectively.
The score of this check will then be 75% = 1 - 3 / 12.
Which means that in 75% of time no bivariate correlation outliers occur.</p>
<div class="ts-check-impact">
<p>A change in the relation between multiple series might indicate potential process upsets,
as well as instrumentation issues.</p>
</div>
"""

from typing import Union
from datetime import timedelta

import pandas as pd
import numpy as np

from pandas.api.types import is_string_dtype

from timeseer import (
    AnalysisInput,
    AnalysisResult,
    BivariateCheckResult,
    DataType,
    MultivariateAnalysisInput,
)
from timeseer.analysis.utils import event_frames_from_dataframe


_CHECK_NAME = "Bivariate correlation outlier"
_EVENT_FRAME_NAME = "Bivariate correlation outlier"
_MIN_SERIES = 2

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": _MIN_SERIES,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [
                DataType.FLOAT32,
                DataType.FLOAT64,
                DataType.DICTIONARY,
                DataType.CATEGORICAL,
            ],
        }
    ],
    "signature": "bivariate",
}


def _get_anomalies(lower, upper, arr):
    a_lower = arr < lower
    a_upper = arr > upper
    return a_lower | a_upper


def _find_anomalous_days(correlations):
    q25, q75 = np.nanquantile(np.array(correlations, dtype=float), [0.25, 0.75], axis=0)
    iqr = q75 - q25
    upper = q75 + 1.5 * iqr
    lower = q25 - 1.5 * iqr
    return [
        _get_anomalies(lower, upper, x) for x in np.array(correlations, dtype=float)
    ]


def _get_intervals(series_anomaly, days):
    anomalies = pd.Series(data=series_anomaly["correlations"], index=days.index)
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()

    intervals = (
        days.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["end_date"] = intervals["end_date"] + timedelta(
        hours=23, minutes=59, seconds=59
    )
    intervals["type"] = _EVENT_FRAME_NAME
    return intervals


def _get_results_per_series(series_anomaly, days) -> BivariateCheckResult:
    intervals = _get_intervals(series_anomaly, days)

    return BivariateCheckResult(
        _CHECK_NAME,
        series_anomaly["series_x"],
        series_anomaly["series_y"],
        event_frames=list(event_frames_from_dataframe(intervals)),
    )


def _anomalous_days_per_series(anomalous_days_matrix, series):
    results = []
    for i in range(len(series) - 1):
        for j in range(i + 1, len(series)):
            results.append(
                {
                    "series_x": series[i],
                    "series_y": series[j],
                    "correlations": [day[i, j] for day in anomalous_days_matrix],
                }
            )
    return results


def _clean_input(inputs) -> pd.DataFrame:
    return (
        pd.concat(
            [
                series.data[~series.data.index.duplicated(keep="first")]["value"]
                for series in inputs
            ],
            axis=1,
            sort=False,
        )
        .interpolate("time")
        .dropna()
        .sort_index()
    )


def _run_correlation_outlier_detection(
    inputs: list[AnalysisInput],
) -> Union[list[BivariateCheckResult], None]:
    concatenated_df = _clean_input(inputs)
    if len(concatenated_df) == 0:
        return None
    concatenated_df = concatenated_df[(concatenated_df != 0).all(1)]

    daily_correlations = np.array(
        [x[1].corr() for x in concatenated_df.resample("D")], dtype=object
    )

    if len(daily_correlations) == 0:
        return None

    anomalous_days_matrix = _find_anomalous_days(daily_correlations)

    series = [series.metadata.series for series in inputs]
    anomalous_days_per_series = _anomalous_days_per_series(
        anomalous_days_matrix, series
    )

    days = concatenated_df.resample("D").mean()

    return [
        _get_results_per_series(series_anomaly, days)
        for series_anomaly in anomalous_days_per_series
    ]


def _filter_invalid_inputs(
    inputs: list[AnalysisInput],
) -> list[AnalysisInput]:
    valid_inputs = []
    for check_input in inputs:
        if is_string_dtype(check_input.data["value"]):
            continue
        valid_inputs.append(check_input)
    return valid_inputs


def run(  # pylint: disable=missing-function-docstring
    analysis_input: MultivariateAnalysisInput,
) -> AnalysisResult:
    inputs = _filter_invalid_inputs(analysis_input.inputs)
    if len(inputs) < _MIN_SERIES:
        return AnalysisResult()

    results = _run_correlation_outlier_detection(inputs)
    if results is None:
        return AnalysisResult()

    return AnalysisResult(bivariate_check_results=results)



"""Identification of a difference in consecutive measurements bigger than a given threshold.
"""


from datetime import datetime
import pandas as pd

from scipy.ndimage import shift

from timeseer import AnalysisInput, AnalysisResult, EventFrame

from timeseer import DataType
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    process_open_intervals,
    handle_open_intervals,
)


META = {
    "checks": [
        {
            "name": "Jump up bigger than threshold",
            "event_frames": ["Jump up bigger than threshold"],
        },
        {
            "name": "Jump down bigger than threshold",
            "event_frames": ["Jump down bigger than threshold"],
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 2,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _get_relevant_jumps(values, threshold):
    down_jumps = [False] * len(values)
    up_jumps = [False] * len(values)

    up_jumps = values > threshold
    up_jumps = up_jumps | shift(up_jumps, -1, cval=False)

    down_jumps = values < -threshold
    down_jumps = down_jumps | shift(down_jumps, -1, cval=False)

    return up_jumps, down_jumps


def _get_intervals(outliers, df, event_type):
    if outliers is None:
        return pd.DataFrame()
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_jump_check(
    analysis_input: AnalysisInput,
) -> tuple[list[EventFrame], datetime]:
    df = _clean_dataframe(analysis_input.data)
    threshold = analysis_input.parameters["threshold"]

    value_diff = df["value"].diff()

    active_points_upwards, active_points_downwards = _get_relevant_jumps(
        value_diff, threshold
    )

    intervals_downwards = _get_intervals(
        active_points_downwards, df, "Jump down bigger than threshold"
    )
    intervals_downwards = handle_open_intervals(df, intervals_downwards)

    intervals_upwards = _get_intervals(
        active_points_upwards, df, "Jump up bigger than threshold"
    )
    intervals_upwards = handle_open_intervals(df, intervals_upwards)

    all_intervals = pd.concat([intervals_downwards, intervals_upwards])

    frames = event_frames_from_dataframe(process_open_intervals(all_intervals))

    last_analyzed_point = df.index[-2]

    return list(frames), last_analyzed_point


def _get_relevant_statistic(analysis_input, stat_name):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return statistics[0]


def _is_valid_input(analysis_input, median_archival_step: float) -> tuple[str, bool]:
    if median_archival_step is None:
        return "No median archival step", False
    if "threshold" not in analysis_input.parameters:
        return "No threshold parameter provided", False
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No clean data", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    median_archival_step = _get_relevant_statistic(
        analysis_input, "Archival time median"
    )
    message, is_ok = _is_valid_input(analysis_input, median_archival_step)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    event_frames, last_analyzed_point = _run_jump_check(
        analysis_input,
    )

    return AnalysisResult(
        event_frames=event_frames,
        last_analyzed_point=last_analyzed_point,
    )



import sqlite3

from yoyo import step

__depends__ = {"008__rename_default_kpi_set"}


def rename_default_kpi_set(db: sqlite3.Connection):
    cursor = db.cursor()
    cursor.execute(
        """
            update KPISets set origin = 'default' where origin in ('control loop', 'preset')
        """
    )
    cursor.execute(
        """
            update KPIS set origin = 'default' where origin in ('control loop', 'preset')
        """
    )


steps = [
    step(rename_default_kpi_set),
]



"""Identification of downtime

<p>This check tries to automatically detect periods of downtime for a series set within a specified time-frame.
</p>
<p><img src='../static/images/reporting/downtime.svg'></p>
<div class="ts-check-impact">
<p>Automated analysis flows typically have difficulties during periods in which they
are not intended to run. Automatically flaggin periods of downtime can build in conditional
execution of these flows.</p>
</div>
"""

from datetime import timedelta
from typing import Optional

import pandas as pd
import numpy as np

from pandas.api.types import is_string_dtype

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import event_frames_from_dataframe
from timeseer.metadata import fields


_CHECK_NAME = "Downtime"
_EVENT_FRAME_NAME = "Downtime"
_MIN_SERIES = 2

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": _MIN_SERIES,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "multivariate",
}


def _is_frame_long_enough(frame):
    return (frame.end_date - frame.start_date) >= timedelta(hours=2)


def _filter_event_frames(all_frames):
    filter_iterator = filter(_is_frame_long_enough, all_frames)
    return filter_iterator


def _get_event_frames(df, anomalies):
    anomalies = pd.Series(data=anomalies, index=df.index)
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()

    intervals = (
        df.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = _EVENT_FRAME_NAME

    return event_frames_from_dataframe(intervals)


def _combine_downtime_indicators(anomalies):
    arr_anomalies = np.array(anomalies)
    return np.sum(arr_anomalies, axis=0, dtype=bool)


def _identify_long_gaps_in_data(df, analysis_input):
    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]
    if median_archival_step is None or len(median_archival_step) == 0:
        return pd.Series([False] * len(df))

    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()

    gaps = diff_times >= 3 * timedelta(seconds=median_archival_step[0])
    return gaps


def _is_dense(values):
    return sum(values) / len(values) > 0.5


def _get_extreme_cutoffs(
    analysis_input: timeseer.AnalysisInput,
) -> tuple[Optional[float], Optional[float]]:
    lower = analysis_input.metadata.get_field(fields.LimitLowFunctional)
    upper = analysis_input.metadata.get_field(fields.LimitHighFunctional)

    if analysis_input.calculated_metadata is not None:
        calculated_lower = analysis_input.calculated_metadata.get_field(
            fields.LimitLowFunctional
        )
        calculated_upper = analysis_input.calculated_metadata.get_field(
            fields.LimitHighFunctional
        )

        if calculated_lower is not None:
            lower = calculated_lower
        if calculated_upper is not None:
            upper = calculated_upper

    return lower, upper


def _clean_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _identify_high_density_of_extremes(
    df: pd.DataFrame, analysis_input: timeseer.AnalysisInput
):
    df = _clean_dataframe(df)
    if len(df) == 0:
        return pd.Series([False] * len(df))

    lower, upper = _get_extreme_cutoffs(analysis_input)
    if lower is None or upper is None:
        return pd.Series([False] * len(df))
    value_range = upper - lower

    anomalies = (df <= lower - value_range) | (df >= upper + value_range)
    sums = anomalies.rolling("1H").sum().fillna(0)
    counts = anomalies.rolling("1H").count().fillna(1)
    return (sums / counts) >= 0.5


def _get_low_value(analysis_input: timeseer.AnalysisInput) -> Optional[float]:
    limit_low = analysis_input.metadata.get_field(fields.LimitLowFunctional)
    if limit_low is not None:
        return limit_low
    if analysis_input.calculated_metadata is not None:
        calculated_limit_low = analysis_input.calculated_metadata.get_field(
            fields.LimitLowFunctional
        )
        if calculated_limit_low is not None:
            return calculated_limit_low
    return None


def _identify_close_to_low(df: pd.DataFrame, analysis_input: timeseer.AnalysisInput):
    low = _get_low_value(analysis_input)
    if low is None:
        return pd.Series([False] * len(df))

    low_values = df.to_numpy() <= (low + low / 100)
    return low_values


def _run_single_downtime_detection(
    clean_df: pd.DataFrame, analysis_input: timeseer.AnalysisInput
):
    lows = _identify_close_to_low(clean_df, analysis_input)
    extremes = _identify_high_density_of_extremes(clean_df, analysis_input)
    gaps = _identify_long_gaps_in_data(clean_df, analysis_input)

    return _combine_downtime_indicators([lows, extremes, gaps])


def _significant_downtime(all_anomalies):
    arr_anomalies = np.array(all_anomalies)
    nb_series = arr_anomalies.shape[0]
    return np.sum(arr_anomalies, axis=0) >= 0.25 * nb_series


def _run_downtime_detection(
    inputs: list[timeseer.AnalysisInput],
) -> list[timeseer.EventFrame]:
    concatenated_df = (
        pd.concat(
            [
                series.data[~series.data.index.duplicated(keep="first")]
                for series in inputs
            ],
            axis=1,
            sort=False,
        )
        .interpolate("time")
        .dropna()
    )
    all_anomalies = [
        _run_single_downtime_detection(concatenated_df.iloc[:, i], analysis)
        for i, analysis in enumerate(inputs)
    ]

    combined_anomalies = _significant_downtime(all_anomalies)

    return list(
        _filter_event_frames(_get_event_frames(concatenated_df, combined_anomalies))
    )


def _filter_invalid_inputs(
    inputs: list[timeseer.AnalysisInput],
) -> list[timeseer.AnalysisInput]:
    valid_inputs = []
    for check_input in inputs:
        if is_string_dtype(check_input.data["value"]):
            continue
        valid_inputs.append(check_input)
    return valid_inputs


def run(
    analysis_input: timeseer.MultivariateAnalysisInput,
):  # pylint: disable=missing-function-docstring
    inputs = _filter_invalid_inputs(analysis_input.inputs)
    if len(inputs) < _MIN_SERIES:
        return timeseer.AnalysisResult()

    event_frames = _run_downtime_detection(inputs)

    return timeseer.AnalysisResult(event_frames=event_frames)



"""There should be no NaN values in the data.

<p>It is possible for values to be recorded in a non-numerical format due to preprocessing or network issues.
These values are indicated as NaN (Not-a-Number).
Non-numerical reading of a sensor or dividing by zero in preprocessing are typical examples of this issue.</p>
<p><img src='../static/images/reporting/nan.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where the corresponding value is not defined (i.e. NaN).
Imagine that 100 points are analyzed in a given time-frame
and there are 10 points whose value is NaN.
The score for this check in that case would be
90% = 1 - 10 / 100. Which means that 90% of all points have valid values.</p>
<div class="ts-check-impact">
<p>
NaNs can hamper (automated) downstream analytics.
</p>
</div>
"""

import pandas as pd
import numpy as np

from timeseer import AnalysisInput, AnalysisResult, EventFrame

from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    handle_open_intervals,
    process_open_intervals,
)

_CHECK_NAME = "NaNs"

META: dict = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": ["NaN"],
        }
    ],
    "conditions": [{"min_series": 1, "min_data_points": 1}],
    "signature": "univariate",
}


def _get_intervals(
    outliers: pd.Series, df: pd.DataFrame, event_type: str
) -> pd.DataFrame:
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _get_active_points(df: pd.DataFrame) -> pd.Series:
    return df["value"].isna()


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].sort_index()


def _run_nan_check(df: pd.DataFrame) -> list[EventFrame]:
    df = _clean_dataframe(df)

    active_points = _get_active_points(df)
    intervals = _get_intervals(active_points, df, "NaN")
    intervals = handle_open_intervals(df, intervals)
    intervals = process_open_intervals(intervals)

    frames = event_frames_from_dataframe(intervals)

    if len(df) == 1:
        if np.isnan(df["value"][0]):
            return list(frames)
        return []

    return list(frames)


def _is_valid_input(median_archival_step: list[float]) -> tuple[str, bool]:
    if median_archival_step is None or len(median_archival_step) == 0:
        return "No median archival step", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(analysis_input: AnalysisInput) -> AnalysisResult:
    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]
    message, is_ok = _is_valid_input(median_archival_step)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    frames = _run_nan_check(analysis_input.data)
    return AnalysisResult(
        event_frames=frames, last_analyzed_point=analysis_input.data.index[-1]
    )



"""There is a period between data being recorded longer than a user-specified number of seconds.
"""

from datetime import datetime, timedelta

import numpy as np
import pandas as pd

from timeseer import AnalysisInput, AnalysisResult, EventFrame

from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    process_open_intervals,
    handle_open_intervals,
)

_CHECK_NAME = "Archival gap longer than threshold"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": ["Archival gap longer than threshold"],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 2,
        }
    ],
    "signature": "univariate",
}


def _get_last_analyzed_point(df: pd.DataFrame, intervals: pd.Series):
    if len(intervals) == 0:
        return df.index[-1]

    if intervals.iloc[-1]["end_date"] < df.index[-1]:
        return df.index[-1]

    points_before_last_analysis = df.index[df.index <= intervals.iloc[-1]["start_date"]]
    return points_before_last_analysis[-1]


def _is_frame_long_enough(frame, df, delta):
    end_date = frame.end_date
    if frame.end_date is None:
        end_date = df.index[-1]

    return (
        end_date.replace(tzinfo=None) - frame.start_date.replace(tzinfo=None)
    ) >= timedelta(seconds=delta)


def _filter_stale_event_frames(all_frames, df, delta):
    filter_iterator = filter(lambda x: _is_frame_long_enough(x, df, delta), all_frames)
    return filter_iterator


def _get_intervals(active_points, df, event_type):
    interval_grp = (active_points != active_points.shift().bfill()).cumsum()
    active_points[active_points.isna()] = 0
    active_points = np.array(active_points, dtype=bool)
    intervals = (
        df.assign(interval_grp=interval_grp)[active_points]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = event_type
    return intervals


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].sort_index()


def _run_archival_gap_longer_than_threshold(
    analysis_input: AnalysisInput,
) -> tuple[list[EventFrame], datetime]:
    df = _clean_dataframe(analysis_input.data)

    delta = analysis_input.parameters["threshold"]
    active_points = pd.DataFrame(
        {"value": ((df.index.to_series().diff()).dt.total_seconds() > delta)}
    )
    active_points = active_points | active_points.shift(-1).bfill()

    intervals = _get_intervals(active_points, df, _CHECK_NAME)
    intervals = handle_open_intervals(df, intervals)

    frames = _filter_stale_event_frames(
        event_frames_from_dataframe(process_open_intervals(intervals)), df, delta
    )

    last_analyzed_point = _get_last_analyzed_point(df, intervals)

    return frames, last_analyzed_point


def _get_relevant_statistic(analysis_input, stat_name):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return statistics[0]


def _is_valid_input(
    analysis_input: AnalysisInput, median_archival_step: float
) -> tuple[str, bool]:
    if median_archival_step is None:
        return "No median archival step", False
    if len(analysis_input.data) == 0:
        return "No data", False
    if "threshold" not in analysis_input.parameters:
        return "No threshold parameter provided", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    median_archival_step = _get_relevant_statistic(
        analysis_input, "Archival time median"
    )
    message, is_ok = _is_valid_input(analysis_input, median_archival_step)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    frames, last_analyzed_point = _run_archival_gap_longer_than_threshold(
        analysis_input
    )

    return AnalysisResult(
        event_frames=list(frames),
        last_analyzed_point=last_analyzed_point,
    )



"""The last archived timestamp of a series."""

import timeseer

_STATISTIC_NAME = "Archival time: latest"

META = {
    "statistics": [{"name": _STATISTIC_NAME}],
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
        }
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:

    last_timestamp = analysis_input.data.index.to_series()[-1].to_pydatetime()
    return timeseer.AnalysisResult(
        statistics=[timeseer.Statistic(_STATISTIC_NAME, "datetime", last_timestamp)]
    )



"""Identification of drift between 2 series that should have a linear relation by means of cusum.

<p>This check identifies periods where the values from sensors that are in a linear relationship
start to drift away from their expected values.</p>
<p><img src='../static/images/reporting/sensor_drift.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where  drift is identified. Imagine that 100 points are analyzed in a given time-frame
and that drift is detected for 10 (consecutive) points. The score for this check in that case would be
90% = 1 - 10 / 100. Which means that for 90% of all points no drift occurs.</p>
<div class="ts-check-impact">
<p>Changes in a physical relation between a set of series could indicate process or instrumentation issues.</p>
</div>
"""

import pandas as pd
import numpy as np


from timeseer import (
    AnalysisInput,
    AnalysisResult,
    BivariateCheckResult,
    DataType,
    MultivariateAnalysisInput,
)
from timeseer.analysis.utils import event_frames_from_dataframe


_CHECK_NAME = "redundant sensor drift (smape)"
_EVENT_FRAME_NAME = "redundant sensor drift (smape)"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": 2,
            "max_series": 2,
        },
        {
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        },
    ],
    "signature": "bivariate",
}


def _clean_input(inputs):
    return (
        pd.concat(
            [
                series.data["value"][
                    ~series.data["value"].index.duplicated(keep="first")
                ]
                for series in inputs
            ],
            axis=1,
            sort=False,
        )
        .interpolate("time")
        .dropna()
        .sort_index()
    )


def _get_smape(concatenated_df):
    series_a = concatenated_df.iloc[:, 0].values
    series_b = concatenated_df.iloc[:, 1].values
    top = np.abs(series_a - series_b)
    denom = np.abs(series_a) + np.abs(series_b)
    return np.divide(top, denom, out=np.zeros_like(top), where=denom != 0)


def _cusum(drift_series, likelihood_function, anomaly_threshold):
    smape = drift_series[0]["drifts"][0]
    s_high = np.zeros_like(smape)
    s_low = np.zeros_like(smape)
    for i in np.arange(1, len(smape)):
        s_high[i] = np.max([0, smape[i] - likelihood_function + s_high[i - 1]])
        s_low[i] = np.max([0, -likelihood_function - smape[i] + s_low[i - 1]])
    anomalies = (s_high > anomaly_threshold) | (s_low > anomaly_threshold)
    return anomalies


def _get_intervals(anomalies, df):
    anomalies = pd.Series(data=anomalies, index=df.index)
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()

    intervals = (
        df.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = _EVENT_FRAME_NAME
    return intervals


def _get_result_per_series(series, df) -> BivariateCheckResult:
    anomalies = _cusum(series, 0.05, 0.2)
    intervals = _get_intervals(anomalies, df)
    frames = event_frames_from_dataframe(intervals)

    return BivariateCheckResult(
        _CHECK_NAME,
        series[0]["series_x"],
        series[0]["series_y"],
        event_frames=list(frames),
    )


def _get_individual_difference_series(error, series):
    results = []
    results.append({"series_x": series[0], "series_y": series[1], "drifts": [error]})
    return results


def _is_valid_input(analysis_input: MultivariateAnalysisInput) -> tuple[str, bool]:
    if len(_clean_input(analysis_input.inputs)) == 0:
        return "No data", False
    return "OK", True


def _run_detect_redundant_sensor_cusum(
    inputs: list[AnalysisInput],
) -> BivariateCheckResult:
    concatenated_df = _clean_input(inputs)

    error = _get_smape(concatenated_df)

    series = [series.metadata.series for series in inputs]
    individual_difference_series = _get_individual_difference_series(error, series)

    return _get_result_per_series(individual_difference_series, concatenated_df)


def run(
    analysis_input: MultivariateAnalysisInput,
):  # pylint: disable=missing-function-docstring

    message, is_ok = _is_valid_input(analysis_input)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    inputs = analysis_input.inputs

    results = _run_detect_redundant_sensor_cusum(inputs)

    return AnalysisResult(bivariate_check_results=[results])



"""For certain units (e.g. pH) there is a limited range values can take.

<p>This check validates whether for a specific set of units, the configured limits fall
within the expected deterministic range.</p>
<p class="scoring-explanation">The score for this checks is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>Mismatches between unit range and physical limits are data hygiene bugs.
</p>
</div>
"""

from enum import Enum, auto
from typing import Optional

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.analysis.utils import (
    get_dimension,
    get_symbol,
    get_limited_units,
    get_unit_range,
)
from timeseer.metadata import fields

_FUNCTIONAL_CHECK_NAME = "Limits inconsistent with unit (functional)"
_PHYSICAL_CHECK_NAME = "Limits inconsistent with unit (physical)"
_FUNCTIONAL_EVENT_FRAME_NAME = "Limits inconsistent with unit (functional)"
_PHYSICAL_EVENT_FRAME_NAME = "Limits inconsistent with unit (physical)"

META = {
    "checks": [
        {
            "name": _FUNCTIONAL_CHECK_NAME,
            "group": "Unit - Functional",
            "event_frames": [_FUNCTIONAL_EVENT_FRAME_NAME],
            "data_type": "bool",
        },
        {
            "name": _PHYSICAL_CHECK_NAME,
            "group": "Unit - Physical",
            "event_frames": [_PHYSICAL_EVENT_FRAME_NAME],
            "data_type": "bool",
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


class LimitType(Enum):
    """Choose whether to use physical or functional limits."""

    PHYSICAL = auto()
    FUNCTIONAL = auto()


def _limit_settings_outside_range(
    analysis_input: AnalysisInput,
    limit_type: LimitType,
    min_value: float,
    max_value: float,
) -> Optional[bool]:
    if limit_type == LimitType.PHYSICAL:
        limit_low = analysis_input.metadata.get_field(fields.LimitLowPhysical)
        limit_high = analysis_input.metadata.get_field(fields.LimitHighPhysical)
    else:
        limit_low = analysis_input.metadata.get_field(fields.LimitLowFunctional)
        limit_high = analysis_input.metadata.get_field(fields.LimitHighFunctional)
    if limit_low and limit_low < min_value:
        return True
    if limit_high and limit_high > max_value:
        return True
    if not (limit_low or limit_high):
        return None
    return False


def _run_values_outside_unit_range_check(
    unit, analysis_input: AnalysisInput, limit_type: LimitType
) -> Optional[bool]:
    min_value, max_value = get_unit_range(unit)
    return _limit_settings_outside_range(
        analysis_input, limit_type, min_value=min_value, max_value=max_value
    )


def _is_valid_input(unit: str) -> tuple[str, bool]:
    if unit is None:
        return "No unit", False
    units_zero_min, dimensions_zero_min = get_limited_units()
    if (get_dimension(unit) not in dimensions_zero_min) & (
        get_symbol(unit) not in units_zero_min
    ):
        return "Unit range is unknown", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    unit = analysis_input.metadata.get_field(fields.Unit)
    message, is_ok = _is_valid_input(unit)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    event_frames = []
    functional_score = _run_values_outside_unit_range_check(
        unit, analysis_input, LimitType.FUNCTIONAL
    )
    if functional_score is not None and functional_score is True:
        event_frames.append(
            EventFrame(
                type=_FUNCTIONAL_EVENT_FRAME_NAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=None,
            )
        )

    physical_score = _run_values_outside_unit_range_check(
        unit, analysis_input, LimitType.PHYSICAL
    )

    if physical_score is not None and physical_score is True:
        event_frames.append(
            EventFrame(
                type=_PHYSICAL_EVENT_FRAME_NAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=None,
            )
        )

    return AnalysisResult(event_frames=event_frames)



"""Identify tags that drift continuously on daily average for at least a month.

<p>A common indication of sensor issues is sensor drift. This is where a measurement of
a sensor starts to slowly deviate from the real measurement. This can be identified in
reduntant sensor cases where for instance the difference between two identical sensors slowly
increases over time. A drift in a measurement can also have process causes
(e.g. fouling, catalyst degradation)</p>
<p><img src='../static/images/reporting/daily_drift.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the sum of time of all
the event-frames containing positive and negative slopes. E.g. assume a total period being analyzed of 1 year and
2 event-frames, one negative and one positive, of 1 month and 2 months respectively.
The score of this check will then be 75% = 1 - 3 / 12. Which means that in 75% of time no drift is detected.</p>
<div class="ts-check-impact">
<p>
Sensor drift can lead to increased costs and variability in the process.
Typically drift is detected manually after a long period of time, which can cause slow deterioration
of product quality.
The longer drift goes unnoticed, the higher the cost and often also the higher the consequences.
Early drift detection can turn a maintenance action into a cleaning action.
Any type of anomaly detection based on normal behavior of the phyiscal ensor can act as a guide for
prioritization of callibration / maintenance.
</p>
</div>
"""

from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import pymannkendall as mk

from pandas.api.types import is_string_dtype

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    process_open_intervals,
    handle_open_intervals,
)

_CHECK_NAME = "Daily drift"


META = {
    "checks": [
        {
            "name": "Daily drift (upwards)",
            "event_frames": ["Positive slopes"],
        },
        {
            "name": "Daily drift (downwards)",
            "event_frames": ["Negative slopes"],
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 4,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _merge_frames(frames: pd.DataFrame) -> pd.DataFrame:
    if len(frames) == 0:
        return frames

    merged_frame = pd.DataFrame(
        {
            "start_date": pd.Series(dtype="datetime64[ns, UTC]"),
            "end_date": pd.Series(dtype="datetime64[ns, UTC]"),
            "type": [],
            "open": [],
        }
    )

    frames = frames.sort_values(by=["start_date"])

    current_start = frames.iloc[0]["start_date"]
    current_end = frames.iloc[0]["end_date"]
    current_type = frames.iloc[0]["type"]
    current_open = frames.iloc[0]["open"]
    for index, row in frames.reset_index(drop=True).iterrows():
        if current_end < row["start_date"]:
            merged_frame = pd.concat(
                [
                    merged_frame,
                    pd.DataFrame(
                        {
                            "start_date": [current_start],
                            "end_date": [current_end],
                            "type": [current_type],
                            "open": [current_open],
                        }
                    ),
                ]
            )
            current_start = row["start_date"]
        if index == len(frames) - 1:
            if current_end < row["start_date"]:
                merged_frame = pd.concat(
                    [
                        merged_frame,
                        pd.DataFrame(
                            {
                                "start_date": [row["start_date"]],
                                "end_date": [row["end_date"]],
                                "type": [row["type"]],
                                "open": [row["open"]],
                            }
                        ),
                    ]
                )
            else:
                merged_frame = pd.concat(
                    [
                        merged_frame,
                        pd.DataFrame(
                            {
                                "start_date": [current_start],
                                "end_date": [row["end_date"]],
                                "type": [current_type],
                                "open": row["open"],
                            }
                        ),
                    ]
                )

        current_end = row["end_date"]
        current_type = row["type"]
        current_open = row["open"]
    merged_frame["open"] = merged_frame["open"].astype("bool")
    return merged_frame


def _get_intervals(df, anomalies, interval_type):
    anomaly_grp = (anomalies != anomalies.shift().bfill()).cumsum()
    intervals = (
        df.assign(anomaly_grp=anomaly_grp)[anomalies]
        .reset_index()
        .groupby(["anomaly_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = interval_type
    intervals["start_date"] = intervals["start_date"] - timedelta(days=27)
    return intervals


def _get_global_outliers(df):
    q25, q75 = np.quantile(df, [0.25, 0.75])
    iqr = q75 - q25
    upper = q75 + 1.5 * iqr
    lower = q25 - 1.5 * iqr
    return lower, upper


def _apply_mk(df, lower, upper):
    if any(df > upper) or any(df < lower):
        return 0

    values = df.values
    q25, q75 = np.quantile(values, [0.25, 0.75])
    iqr = q75 - q25
    values[(values > q75 + 1.5 * iqr) | (values < q25 - 1.5 * iqr)] = np.nan
    values = values[~np.isnan(values)]
    values = values[np.insert(np.diff(values).astype(bool), 0, True)]

    if len(values) <= 20:
        return 0
    result = mk.original_test(values, alpha=0.001)
    if result.trend == "no trend":
        return 0
    if result.trend == "increasing":
        return 1
    return -1


def _get_last_analyzed_point(df):
    points_before_last_analysis = df.index[
        df.index <= df.index[-1] - timedelta(weeks=4)
    ]
    if len(points_before_last_analysis) == 0:
        return df.index[0]
    return points_before_last_analysis[-1]


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_drift_check(  # pylint: disable=too-many-locals
    analysis_input: AnalysisInput,
) -> tuple[list[EventFrame], datetime]:
    df = _clean_dataframe(analysis_input.data)

    df_daily = df.resample("1D").median().interpolate("time")
    lower, upper = _get_global_outliers(df_daily)
    slopes = df_daily.rolling(28).apply(lambda x: _apply_mk(x, lower, upper))
    pos_slopes = slopes["value"] == 1
    neg_slopes = slopes["value"] == -1

    pos_intervals = _get_intervals(df_daily, pos_slopes, "Positive slopes")
    neg_intervals = _get_intervals(df_daily, neg_slopes, "Negative slopes")
    pos_intervals = handle_open_intervals(df_daily, pos_intervals)
    neg_intervals = handle_open_intervals(df_daily, neg_intervals)

    mpos = _merge_frames(pos_intervals)
    mneg = _merge_frames(neg_intervals)

    all_intervals = pd.concat((mpos, mneg))

    all_frames = event_frames_from_dataframe(process_open_intervals(all_intervals))

    last_analyzed_point = _get_last_analyzed_point(df)

    return list(all_frames), last_analyzed_point


def _is_valid_input(data: pd.DataFrame) -> tuple[str, bool]:
    df = _clean_dataframe(data)
    if len(df) == 0:
        return "Not enough data", False
    if len(df["value"].unique()) <= 3:
        return "Not enough unique values", False
    if is_string_dtype(df["value"]):
        return "Can not be a string", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(analysis_input: AnalysisInput) -> AnalysisResult:
    message, is_ok = _is_valid_input(analysis_input.data)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    frames, last_analyzed_point = _run_drift_check(analysis_input)
    return AnalysisResult(event_frames=frames, last_analyzed_point=last_analyzed_point)



import sqlite3

from yoyo import step

__depends__ = {"006__contribution_evaluations"}


def migrate_data_service_not_null_constraints(db: sqlite3.Connection):
    cursor = db.cursor()
    cursor.execute("pragma schema_version")
    (schema_version,) = cursor.fetchone()
    cursor.execute("pragma writable_schema = ON")
    cursor.execute(
        """
        update sqlite_master
        set sql = 'create table DataServiceBadActorScore(
    id integer primary key autoincrement,
    data_service_view_evaluation_id integer not null,
    series_id text,
    start_date datetime not null,
    end_date datetime not null,
    state text not null,
    score real not null,

    foreign key (data_service_view_evaluation_id) references DataServiceViewEvaluations(id) on delete cascade
)'
        where type = 'table' and name = 'DataServiceBadActorScore'
    """
    )
    cursor.execute(
        """
        update sqlite_master
        set sql = 'CREATE TABLE Scores(
    id integer primary key autoincrement,
    data_service_view_evaluation_id integer not null,
    series_id text,
    score_name text not null,
    score float not null,

    foreign key (data_service_view_evaluation_id) references DataServiceViewEvaluations(id) on delete cascade
)'
        where type = 'table' and name = 'Scores'
    """
    )
    cursor.execute(f"pragma schema_version = {schema_version + 1}")
    cursor.execute("pragma writable_schema = OFF")


steps = [
    step(migrate_data_service_not_null_constraints),
]



"""Regime detection event frames.

<p></p>"""

from datetime import timedelta
from typing import Optional

import pandas as pd
import numpy as np

from pandas.api.types import is_string_dtype

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import event_frames_from_dataframe
from timeseer.metadata import fields

_EVENT_FRAME_NAME = "Downtime"

META = {
    "event_frames": [_EVENT_FRAME_NAME],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 2,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _is_frame_long_enough(frame):
    return (frame.end_date - frame.start_date) >= timedelta(hours=2)


def _filter_event_frames(all_frames):
    filter_iterator = filter(_is_frame_long_enough, all_frames)
    return filter_iterator


def _get_event_frames(df, anomalies):
    anomalies = pd.Series(data=anomalies, index=df.index)
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()

    intervals = (
        df.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = _EVENT_FRAME_NAME

    return event_frames_from_dataframe(intervals)


def _combine_downtime_indicators(anomalies):
    arr_anomalies = np.array(anomalies)
    return np.sum(arr_anomalies, axis=0, dtype=bool)


def _identify_long_gaps_in_data(
    df: pd.DataFrame, analysis_input: timeseer.AnalysisInput
):
    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]

    if (
        median_archival_step is None
        or len(median_archival_step) == 0
        or all(np.isnan(median_archival_step))
    ):
        return pd.Series([False] * len(df))

    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()
    gaps = diff_times >= 3 * timedelta(seconds=median_archival_step[0])
    return gaps


def _get_low_value(analysis_input: timeseer.AnalysisInput) -> Optional[float]:
    limit_low = analysis_input.metadata.get_field(fields.LimitLowFunctional)
    if limit_low is not None:
        return limit_low
    if analysis_input.calculated_metadata is not None:
        calculated_limit_low = analysis_input.calculated_metadata.get_field(
            fields.LimitLowFunctional
        )
        if calculated_limit_low is not None:
            return calculated_limit_low
    return None


def _identify_close_to_low(df: pd.DataFrame, analysis_input: timeseer.AnalysisInput):
    low = _get_low_value(analysis_input)
    if low is None:
        return pd.Series([False] * len(df))

    low_values = df["value"] <= low + low / 100
    return low_values


def _get_clean_df(analysis_input: timeseer.AnalysisInput) -> Optional[pd.DataFrame]:
    df = analysis_input.data
    df = df[~df.index.duplicated(keep="first")].dropna().sort_index()
    if len(df) == 0:
        return None
    return df


def _run_downtime_detection(
    analysis_input: timeseer.AnalysisInput,
) -> Optional[list[timeseer.EventFrame]]:
    clean_df = _get_clean_df(analysis_input)
    if clean_df is None:
        return None

    lows = _identify_close_to_low(clean_df, analysis_input)
    gaps = _identify_long_gaps_in_data(clean_df, analysis_input)

    combined_anomalies = _combine_downtime_indicators([lows, gaps])

    return list(_filter_event_frames(_get_event_frames(clean_df, combined_anomalies)))


def _is_input_valid(analysis_input: timeseer.AnalysisInput) -> tuple[str, bool]:
    if is_string_dtype(analysis_input.data["value"]):
        return "Can not be a string", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    message, is_ok = _is_input_valid(analysis_input)
    if not is_ok:
        return timeseer.AnalysisResult(condition_message=message)

    frames = _run_downtime_detection(analysis_input)
    if frames is None:
        return timeseer.AnalysisResult()
    return timeseer.AnalysisResult(event_frames=frames)



"""Time series data should have strict increasing timestamps.

<p>Duplicate time stamps can occur because of connection issues, causing the system
to rely on buffered data. Another common cause for duplicates is the daylight saving time
not handled correctly in the source.</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>
Misaligned ordering of timestamps can hamper (automated) downstream analytics.
</p>
</div>
"""

from datetime import timedelta, timezone

from timeseer import AnalysisInput, AnalysisResult, EventFrame

_CHECK_NAME = "No duplicate timestamps"
_EVENT_FRAME_NAME = "Duplicate timestamp"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "data_type": "bool",
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [{"min_series": 1, "min_data_points": 2}],
    "signature": "univariate",
}


def _find_duplicate_timestamps(df):
    duplicates = df.loc[
        df.index.to_series().diff() == timedelta(0)
    ].index.drop_duplicates(keep="last")
    for ts in duplicates.array:
        py_ts = ts.to_pydatetime().replace(tzinfo=timezone.utc)
        yield EventFrame(_EVENT_FRAME_NAME, py_ts, py_ts)


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    duplicate_timestamps = list(_find_duplicate_timestamps(analysis_input.data))
    return AnalysisResult(event_frames=duplicate_timestamps)



"""The unit should match the dimension (or unit) of the property given as input.
input can be given as unit (e.g. 'fahrenheit', 'mm^3/s') or as the name of property
in brackets (e.g. '[temperature]', '[pressure]')

"""


from pint import UnitRegistry
from pint import UndefinedUnitError, DefinitionSyntaxError
from pint.util import UnitsContainer


from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.metadata import fields
from timeseer.analysis.utils import get_unit_registry


_CHECK_NAME = "Unit not matching dimension"
_EVENT_FRAME_NAME = "Unit not matching dimension"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
            "data_type": [DataType.STRING],
        }
    ],
    "signature": "univariate",
}


def _is_valid_input(
    analysis_input: AnalysisInput, ureg: UnitRegistry
) -> tuple[str, bool]:
    if "value" not in analysis_input.parameters:
        return "No value parameter provided", False
    value = analysis_input.parameters["value"]
    try:
        ureg.get_dimensionality(value)
        return "OK", True
    except (
        UndefinedUnitError,
        AttributeError,
        TypeError,
        ValueError,
        KeyError,
        DefinitionSyntaxError,
    ):
        return "invalid reference value", False


def _get_reference_dimension(
    analysis_input: AnalysisInput, ureg: UnitRegistry
) -> UnitsContainer:
    value = analysis_input.parameters["value"]
    return ureg.get_dimensionality(value)


def _get_unit_dimension(
    analysis_input: AnalysisInput, ureg: UnitRegistry
) -> UnitsContainer:
    unit = analysis_input.metadata.get_field(fields.Unit)
    if "%" in unit:
        return ureg.get_dimensionality("[]")
    return ureg.get_dimensionality(unit)


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    ureg = get_unit_registry()
    message, is_ok = _is_valid_input(analysis_input, ureg)
    if not is_ok:
        return AnalysisResult(condition_message=message)
    event_frames = []
    try:
        unit_dimension = _get_unit_dimension(analysis_input, ureg)
        reference_dimension = _get_reference_dimension(analysis_input, ureg)
        check_fail = unit_dimension != reference_dimension
    except (
        UndefinedUnitError,
        AttributeError,
        TypeError,
        ValueError,
        KeyError,
        DefinitionSyntaxError,
    ):
        check_fail = True

    if check_fail:
        event_frames.append(
            EventFrame(
                type=_EVENT_FRAME_NAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=None,
            )
        )

    return AnalysisResult(event_frames=event_frames)



"""
Stiction
"""

from datetime import datetime
import pandas as pd
import numpy as np

from scipy.signal import find_peaks
from scipy.optimize import curve_fit

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.analysis.utils import (
    handle_open_intervals,
)

from timeseer.analysis.utils.oscillations import control_loop_oscillation

_CHECK_NAME = "Stiction"
_EVENT_FRAME_NAME = "Stiction"


META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
            "data_type": [
                DataType.FLOAT64,
                DataType.FLOAT32,
            ],
        }
    ],
    "signature": "univariate",
}


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _is_valid_input(
    analysis_input: AnalysisInput, median_archival_step: list[float]
) -> tuple[str, bool]:
    if median_archival_step is None or len(median_archival_step) == 0:
        return "No median archival", False
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No clean data", False
    return "OK", True


def _is_correct_series(analysis_input: AnalysisInput) -> tuple[str, bool]:
    position = analysis_input.metadata.get_field_by_name("control loop")
    if position == "OP":
        return "OK", True
    return "Only applicable to OP", False


def _peaks_finder(df):
    peaks, _ = find_peaks(df["value"], height=0)
    peak_ts = df.index[peaks]
    peaks_df = pd.DataFrame(data={"start": peak_ts[:-1], "end": peak_ts[1:]})
    peaks_df["length"] = peaks_df["end"] - peaks_df["start"]
    peaks_df["seconds"] = [
        length.total_seconds() for length in list(peaks_df["length"])
    ]
    return peaks_df


def _get_segments(df, peaks_df):
    segments = []
    for _, row in peaks_df.iterrows():
        ptop = df.loc[row["start"] : row["end"]]
        segments.append(ptop)
    return segments


def _sine_shape(X, amplitude, phase, intercept, angular_frq):
    y = (amplitude * np.sin(angular_frq * X + phase)) + intercept
    return np.array(y)


def _tri_shape(X, tri_edge_x, tri_edge_y, slope_1, slope_2):
    y = []
    for point in X:
        if point < tri_edge_x:
            height = slope_1 * point + tri_edge_y - slope_1 * tri_edge_x
        if point >= tri_edge_x:
            height = slope_2 * point + tri_edge_y - slope_2 * tri_edge_x
        y.append(height)
    return np.array(y)


# pylint: disable=unbalanced-tuple-unpacking
def _check_stiction(curve):
    start_time = curve.index[0].timestamp()
    if len(curve) < 5:
        return float("nan")
    X = curve.index.to_series().apply(lambda X: X.timestamp() - start_time)
    w_segment = 2 * np.pi / (X[-1])
    try:
        sin_p, _ = curve_fit(
            _sine_shape,
            X.values,
            curve["value"],
            bounds=(
                [0, 2 * np.pi / 5, -np.Inf, 0.95 * w_segment],
                [np.Inf, 3 * np.pi / 5, np.Inf, 1.05 * w_segment],
            ),
        )
    except RuntimeError:
        return float("nan")
    mse_sin = np.sum((curve["value"] - _sine_shape(X.values, *sin_p)) ** 2)
    try:
        tri_p, _ = curve_fit(
            _tri_shape,
            X.values,
            curve["value"],
            bounds=([0, -np.Inf, -np.Inf, 0], [X[-1], np.min(curve.values), 0, np.Inf]),
        )
    except RuntimeError:
        return float("nan")
    mse_tri = np.sum((curve["value"] - _tri_shape(X.values, *tri_p)) ** 2)
    p_stiction = mse_sin / (mse_sin + mse_tri)
    return p_stiction


def _get_oscillation_eventframes(analysis_input):
    oscillation_results = control_loop_oscillation.run(analysis_input)
    frames = oscillation_results.event_frames
    return frames


def _eventframes_to_intervals(df, event_frames):
    starts = [frame.start_date for frame in event_frames]
    ends = [frame.end_date for frame in event_frames]
    names = [frame.type for frame in event_frames]
    intervals = pd.DataFrame(
        data={"type": names, "start_date": starts, "end_date": ends}
    )
    if len(intervals) > 0:
        if intervals.iloc[-1, -1] is None:
            intervals.iloc[-1, -1] = df.index[-1]
    if len(intervals) > 0:
        if intervals.iloc[-1, -1] is pd.NaT:
            intervals.iloc[-1, -1] = df.index[-1]
    return intervals


# pylint: disable=too-many-locals
def _run_stiction(analysis_input: AnalysisInput) -> tuple[list[EventFrame], datetime]:
    oscillation_eventframes = _get_oscillation_eventframes(analysis_input)
    df = analysis_input.data
    df = _clean_dataframe(df)
    last_analyzed_point = df.index[-1]
    if len(oscillation_eventframes) == 0:
        return [], last_analyzed_point
    intervals = _eventframes_to_intervals(df, oscillation_eventframes)

    result_frames = []

    for index, row in intervals.iterrows():
        signal = df.loc[row["start_date"] : row["end_date"]]
        peaks_df = _peaks_finder(signal)
        if len(peaks_df) < 3:
            continue
        segments = _get_segments(signal, peaks_df)
        stiction_probs = [_check_stiction(segment) for segment in segments]
        if np.all(np.isnan(stiction_probs)):
            continue
        if np.nanmean(stiction_probs) >= 0.6:
            frame = oscillation_eventframes[index]
            frame.type = _EVENT_FRAME_NAME
            result_frames.append(frame)

    result_intervals = _eventframes_to_intervals(df, result_frames)

    result_intervals = handle_open_intervals(df, result_intervals)

    return result_frames, last_analyzed_point


# pylint: disable=missing-function-docstring
def run(analysis_input: AnalysisInput) -> AnalysisResult:
    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]
    message, is_ok = _is_valid_input(analysis_input, median_archival_step)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    message, is_ok = _is_correct_series(analysis_input)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    frames, last_analyzed_point = _run_stiction(analysis_input)

    return AnalysisResult(event_frames=frames, last_analyzed_point=last_analyzed_point)



"""Within a flow all series with the same dimension should have comparable units.

<p>This check identifies how consistent the units are used within each dimension.</p>
<table style="width:100%">
  <tr>
    <th>Unit</th>
    <th>Dimension</th>
  </tr>
  <tr>
    <td>kPa</td>
    <td>Pressure</td>
  </tr>
  <tr>
    <td>kPa</td>
    <td>Pressure</td>
  </tr>
  <tr>
    <td style="background-color:rgba(250,105,106,1)">barg</td>
    <td>Pressure</td>
  </tr>
  <tr>
    <td>C</td>
    <td>Temperature</td>
  </tr>
</table>
<p class="scoring-explanation">The score of this check is calculated based on the consistency
of the units over all dimensions.
Imagine that a flow contains 4 different dimensions and that for one of those dimensions
there is 1 serie with a deviating unit.
In this case the score would be 80% =  4 (dimensions) / 5 (units)</p>
<div class="ts-check-impact">
<p>Not having comparable or consistent units hinders the interpretation over a set of series.
Downstream analysis will also become responsible for the unit conversion or correct normalization
which can lead to errors.</p>
</div>
"""

import pandas as pd

from pint.util import to_units_container

from timeseer import AnalysisInput, AnalysisResult, MultivariateAnalysisInput, Statistic
from timeseer.analysis.utils import (
    get_unit_registry,
    get_dimension,
    get_symbol,
)
from timeseer.metadata import fields


_STATISTIC_NAME = "Unit consistency"

META = {
    "run": "before",
    "statistics": [
        {"name": _STATISTIC_NAME},
    ],
    "conditions": [
        {
            "min_series": 2,
        }
    ],
    "signature": "multivariate",
}


def _run_unit_to_dimension_ration(inputs: list[AnalysisInput]):
    ureg = get_unit_registry()
    units = pd.DataFrame(map(lambda x: x.metadata.get_field(fields.Unit), inputs))
    units.drop_duplicates(inplace=True)
    units.dropna(inplace=True)
    units.replace(to_replace=".*%.*", value="pct", regex=True, inplace=True)
    units.drop_duplicates(inplace=True)
    units = units[0].apply(get_symbol)
    units.drop_duplicates(inplace=True)
    dimensions = units.apply(get_dimension)
    nb_units = len(
        units[
            ~dimensions.isin(["Unknown", to_units_container("", registry=ureg)])
        ].dropna()
    )
    dimensions.drop_duplicates(inplace=True)
    nb_dimensions = len(
        dimensions[
            ~dimensions.isin(["Unknown", to_units_container("", registry=ureg)])
        ].dropna()
    )
    if nb_units == 0:
        return 1
    return nb_dimensions / nb_units


def run(  # pylint: disable=missing-function-docstring
    analysis_input: MultivariateAnalysisInput,
) -> AnalysisResult:
    inputs = analysis_input.inputs

    unit_to_dimension_ratio = _run_unit_to_dimension_ration(inputs)
    statistic = Statistic(_STATISTIC_NAME, "float", float(unit_to_dimension_ratio))

    return AnalysisResult(statistics=[statistic])
